/* Generated by Cython 0.29.22 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "include_dirs": [
            "/opt/hostedtoolcache/Python/3.8.8/x64/lib/python3.8/site-packages/numpy/core/include"
        ],
        "name": "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector",
        "sources": [
            "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py"
        ]
    },
    "module_name": "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_22"
#define CYTHON_HEX_VERSION 0x001D16F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__octobot_trading__exchanges__connectors__cryptofeed_websocket_connector
#define __PYX_HAVE_API__octobot_trading__exchanges__connectors__cryptofeed_websocket_connector
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py",
  "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.pxd",
  "stringsource",
  "octobot_trading/util/initializable.pxd",
  "octobot_trading/exchanges/abstract_exchange.pxd",
  "octobot_trading/exchanges/exchange_config_data.pxd",
  "octobot_trading/exchanges/exchange_manager.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable;
struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange;
struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig;
struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager;
struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info;
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions;
struct __pyx_opt_args_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled;
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_all_currencies;
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_pairs;
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_market_pair;

/* "config_util.pxd":22
 * cpdef bint is_trader_simulator_enabled(dict config)
 * cpdef bint is_currency_enabled(dict config, str currency, bint default_value)
 * cpdef bint is_trade_history_loading_enabled(dict config, bint default=*)             # <<<<<<<<<<<<<<
 * cpdef list get_symbols(dict config, bint enabled_only)
 * cpdef set get_all_currencies(dict config, bint enabled_only=*)
 */
struct __pyx_opt_args_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled {
  int __pyx_n;
  int __pyx_default;
};

/* "config_util.pxd":24
 * cpdef bint is_trade_history_loading_enabled(dict config, bint default=*)
 * cpdef list get_symbols(dict config, bint enabled_only)
 * cpdef set get_all_currencies(dict config, bint enabled_only=*)             # <<<<<<<<<<<<<<
 * cpdef list get_pairs(dict config, str currency, bint enabled_only=*)
 * cpdef tuple get_market_pair(dict config, str currency, bint enabled_only=*)
 */
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_all_currencies {
  int __pyx_n;
  int enabled_only;
};

/* "config_util.pxd":25
 * cpdef list get_symbols(dict config, bint enabled_only)
 * cpdef set get_all_currencies(dict config, bint enabled_only=*)
 * cpdef list get_pairs(dict config, str currency, bint enabled_only=*)             # <<<<<<<<<<<<<<
 * cpdef tuple get_market_pair(dict config, str currency, bint enabled_only=*)
 * cpdef str get_reference_market(dict config)
 */
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_pairs {
  int __pyx_n;
  int enabled_only;
};

/* "config_util.pxd":26
 * cpdef set get_all_currencies(dict config, bint enabled_only=*)
 * cpdef list get_pairs(dict config, str currency, bint enabled_only=*)
 * cpdef tuple get_market_pair(dict config, str currency, bint enabled_only=*)             # <<<<<<<<<<<<<<
 * cpdef str get_reference_market(dict config)
 * 
 */
struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_market_pair {
  int __pyx_n;
  int enabled_only;
};
struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_get_market_status;
struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_parse_timestamp;

/* "octobot_trading/exchanges/abstract_exchange.pxd":39
 * 
 *     # exchange requests
 *     cpdef dict get_market_status(self, str symbol, object price_example=*, bint with_fixer=*)             # <<<<<<<<<<<<<<
 *     cpdef dict get_trade_fee(self, str symbol, object order_type, double quantity, double price, str taker_or_maker)
 *     cpdef dict get_fees(self, str symbol)
 */
struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_get_market_status {
  int __pyx_n;
  PyObject *price_example;
  int with_fixer;
};

/* "octobot_trading/exchanges/abstract_exchange.pxd":62
 *     cpdef dict parse_order_book(self, dict order_book)
 *     cpdef dict parse_order_book_ticker(self, dict order_book_ticker)
 *     cpdef double parse_timestamp(self, dict data_dict, str timestamp_key, object default_value=*, bint ms=*)             # <<<<<<<<<<<<<<
 *     cpdef str parse_currency(self, str currency)
 *     cpdef str parse_order_id(self, dict order)
 */
struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_parse_timestamp {
  int __pyx_n;
  PyObject *default_value;
  int ms;
};
struct __pyx_opt_args_15octobot_trading_9exchanges_27abstract_websocket_exchange_25AbstractWebsocketExchange_initialize;

/* "octobot_trading/exchanges/abstract_websocket_exchange.pxd":40
 *     cpdef object get_exchange_credentials(self)
 *     cpdef object get_book_instance(self, str symbol)
 *     cpdef void initialize(self, list currencies=*, list pairs=*, list time_frames=*, list channels=*)             # <<<<<<<<<<<<<<
 *     cpdef bint _should_authenticate(self)
 *     cpdef int get_max_handled_pair_with_time_frame(self)
 */
struct __pyx_opt_args_15octobot_trading_9exchanges_27abstract_websocket_exchange_25AbstractWebsocketExchange_initialize {
  int __pyx_n;
  PyObject *currencies;
  PyObject *pairs;
  PyObject *time_frames;
  PyObject *channels;
};

/* "initializable.pxd":23
 * It is also use to store creation & fill values of the order """
 * 
 * cdef class Initializable:             # <<<<<<<<<<<<<<
 *     cdef public bint is_initialized
 */
struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable {
  PyObject_HEAD
  int is_initialized;
};


/* "octobot_trading/exchanges/abstract_exchange.pxd":20
 * 
 * 
 * cdef class AbstractExchange(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef public dict config
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange {
  struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable __pyx_base;
  struct __pyx_vtabstruct_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *__pyx_vtab;
  PyObject *config;
  PyObject *logger;
  PyObject *current_account;
  PyObject *trader;
  PyObject *exchange_manager;
  PyObject *connector;
  PyObject *symbols;
  PyObject *time_frames;
  double allowed_time_lag;
  PyObject *name;
};


/* "octobot_trading/exchanges/exchange_config_data.pxd":19
 * cimport octobot_trading.util as util
 * 
 * cdef class ExchangeConfig(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef object _logger
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig {
  struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable __pyx_base;
  struct __pyx_vtabstruct_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *__pyx_vtab;
  PyObject *_logger;
  PyObject *traded_cryptocurrencies;
  PyObject *config;
  PyObject *traded_symbol_pairs;
  PyObject *all_config_symbol_pairs;
  PyObject *traded_time_frames;
  PyObject *real_time_time_frames;
  PyObject *exchange_manager;
};


/* "octobot_trading/exchanges/exchange_manager.pxd":24
 * cimport octobot_trading.util as util
 * 
 * cdef class ExchangeManager(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef public str id
 *     cdef public str exchange_class_string
 */
struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager {
  struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable __pyx_base;
  struct __pyx_vtabstruct_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *__pyx_vtab;
  PyObject *id;
  PyObject *exchange_class_string;
  PyObject *exchange_name;
  PyObject *config;
  PyObject *tentacles_setup_config;
  PyObject *logger;
  PyObject *backtesting;
  PyObject *trader;
  PyObject *client_time_frames;
  PyObject *client_symbols;
  PyObject *trading_modes;
  int rest_only;
  int ignore_config;
  int is_ready;
  int is_simulated;
  int is_backtesting;
  int is_trader_simulated;
  int is_collecting;
  int is_spot_only;
  int is_margin;
  int is_future;
  int is_sandboxed;
  int is_trading;
  int has_websocket;
  int exchange_only;
  int without_auth;
  struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *exchange;
  struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *exchange_web_socket;
  struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *exchange_config;
  PyObject *exchange_symbols_data;
  PyObject *exchange_personal_data;
};


/* "octobot_trading/exchanges/abstract_websocket_exchange.pxd":19
 * cimport octobot_trading.exchanges.exchange_manager as exchange_manager
 * 
 * cdef class AbstractWebsocketExchange:             # <<<<<<<<<<<<<<
 *     cdef public dict config
 *     cdef public dict books
 */
struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange {
  PyObject_HEAD
  struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *__pyx_vtab;
  PyObject *config;
  PyObject *books;
  struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *exchange_manager;
  PyObject *name;
  PyObject *exchange_id;
  PyObject *currencies;
  PyObject *pairs;
  PyObject *time_frames;
  PyObject *channels;
  PyObject *exchange;
  PyObject *client;
  PyObject *logger;
  PyObject *bot_mainloop;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.pxd":19
 * cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
 * 
 * cdef class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):             # <<<<<<<<<<<<<<
 *     cdef public dict callback_by_feed
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector {
  struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange __pyx_base;
  PyObject *callback_by_feed;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":118
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker {
  PyObject_HEAD
  PyObject *__pyx_v_ask;
  PyObject *__pyx_v_bid;
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_timestamp;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":128
 *                                        bid_price=None)
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade {
  PyObject_HEAD
  PyObject *__pyx_v_amount;
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_order_id;
  PyObject *__pyx_v_price;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_side;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_timestamp;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":143
 *                                        }])
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book {
  PyObject_HEAD
  PyObject *__pyx_v_book;
  PyObject *__pyx_v_book_instance;
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_timestamp;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":172
 *         ]
 * 
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,             # <<<<<<<<<<<<<<
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle {
  PyObject_HEAD
  PyObject *__pyx_v_candle;
  PyObject *__pyx_v_close_price;
  PyObject *__pyx_v_closed;
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_high_price;
  PyObject *__pyx_v_interval;
  PyObject *__pyx_v_low_price;
  PyObject *__pyx_v_open_price;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_start;
  PyObject *__pyx_v_stop;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_time_frame;
  PyObject *__pyx_v_timestamp;
  PyObject *__pyx_v_trades;
  PyObject *__pyx_v_volume;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":196
 *                                            candle=candle)
 * 
 *     async def delta(self, feed, symbol, delta, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta {
  PyObject_HEAD
  PyObject *__pyx_v_delta;
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_timestamp;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":199
 *         pass
 * 
 *     async def liquidations(self, feed, symbol, side, leaves_qty, price, order_id, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations {
  PyObject_HEAD
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_leaves_qty;
  PyObject *__pyx_v_order_id;
  PyObject *__pyx_v_price;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_side;
  PyObject *__pyx_v_symbol;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":202
 *         pass
 * 
 *     async def funding(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding {
  PyObject_HEAD
  PyObject *__pyx_v_kwargs;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":205
 *         pass
 * 
 *     async def open_interest(self, feed, symbol, open_interest, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest {
  PyObject_HEAD
  PyObject *__pyx_v_feed;
  PyObject *__pyx_v_open_interest;
  PyObject *__pyx_v_receipt_timestamp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
  PyObject *__pyx_v_symbol;
  PyObject *__pyx_v_timestamp;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":208
 *         pass
 * 
 *     async def volume(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume {
  PyObject_HEAD
  PyObject *__pyx_v_kwargs;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":211
 *         pass
 * 
 *     async def futures_index(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index {
  PyObject_HEAD
  PyObject *__pyx_v_kwargs;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":214
 *         pass
 * 
 *     async def market_info(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Coingecko only
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info {
  PyObject_HEAD
  PyObject *__pyx_v_kwargs;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
};


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":217
 *         pass  # Coingecko only
 * 
 *     async def transactions(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Whale alert only
 * 
 */
struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions {
  PyObject_HEAD
  PyObject *__pyx_v_kwargs;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self;
};



/* "octobot_trading/exchanges/abstract_exchange.pxd":20
 * 
 * 
 * cdef class AbstractExchange(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef public dict config
 * 
 */

struct __pyx_vtabstruct_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange {
  PyObject *(*get_name)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  PyObject *(*get_market_status)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_get_market_status *__pyx_optional_args);
  PyObject *(*get_trade_fee)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, PyObject *, double, double, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_fees)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  double (*get_uniform_timestamp)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, double, int __pyx_skip_dispatch);
  PyObject *(*get_pair_from_exchange)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_pair)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_pair_cryptocurrency)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_split_pair_from_exchange)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  int (*get_rate_limit)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  PyObject *(*get_default_balance)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  int (*authenticated)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  int (*get_max_handled_pair_with_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  PyObject *(*parse_balance)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_trade)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_order)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_ticker)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_ohlcv)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_order_book)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_order_book_ticker)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  double (*parse_timestamp)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_9exchanges_17abstract_exchange_16AbstractExchange_parse_timestamp *__pyx_optional_args);
  PyObject *(*parse_currency)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_order_id)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_status)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_side)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*parse_account)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*clean_recent_trade)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*clean_trade)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*clean_order)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  double (*get_exchange_current_time)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, int __pyx_skip_dispatch);
  PyObject *(*uniformize_candles_if_necessary)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_uniformized_timestamp)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  PY_LONG_LONG (*get_candle_since_timestamp)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*_uniformize_candles_timestamps)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *);
  void (*_uniformize_candle_timestamps)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *);
  void (*log_order_creation_error)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, PyObject *, PyObject *, double, PyObject *, PyObject *, int __pyx_skip_dispatch);
  void (*handle_token_error)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
  int (*is_supported_order_type)(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange *__pyx_vtabptr_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange;


/* "octobot_trading/exchanges/exchange_config_data.pxd":19
 * cimport octobot_trading.util as util
 * 
 * cdef class ExchangeConfig(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef object _logger
 * 
 */

struct __pyx_vtabstruct_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig {
  void (*set_config_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, int __pyx_skip_dispatch);
  void (*set_config_traded_pairs)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, int __pyx_skip_dispatch);
  PyObject *(*get_shortest_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, int __pyx_skip_dispatch);
  PyObject *(*_is_tradable_with_cryptocurrency)(PyObject *, PyObject *);
  void (*_set_config_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *);
  void (*_set_config_traded_pairs)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *);
  void (*_populate_non_wildcard_pairs)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *, PyObject *, int);
  void (*_populate_wildcard_pairs)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *, PyObject *, int);
  PyObject *(*_add_tradable_symbols_from_config)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *, PyObject *);
  PyObject *(*_add_tradable_symbols)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *, PyObject *);
  PyObject *(*_add_tradable_time_frames)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *);
  PyObject *(*_create_wildcard_symbol_list)(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *, PyObject *);
};
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig *__pyx_vtabptr_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig;


/* "octobot_trading/exchanges/exchange_manager.pxd":24
 * cimport octobot_trading.util as util
 * 
 * cdef class ExchangeManager(util.Initializable):             # <<<<<<<<<<<<<<
 *     cdef public str id
 *     cdef public str exchange_class_string
 */

struct __pyx_vtabstruct_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager {
  void (*_load_config_symbols_and_time_frames)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *);
  int (*enabled)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  void (*load_constants)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_symbol)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_quote_and_base)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_rest_pairs_refresh_threshold)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  int (*need_user_stream)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  void (*reset_exchange_symbols_data)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  void (*reset_exchange_personal_data)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  int (*check_config)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  int (*symbol_exists)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  int (*time_frame_exists)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_name)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  int (*get_currently_handled_pair_with_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  int (*get_is_overloaded)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_credentials)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*should_decrypt_token)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_symbol_data)(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *__pyx_vtabptr_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager;


/* "octobot_trading/exchanges/abstract_websocket_exchange.pxd":19
 * cimport octobot_trading.exchanges.exchange_manager as exchange_manager
 * 
 * cdef class AbstractWebsocketExchange:             # <<<<<<<<<<<<<<
 *     cdef public dict config
 *     cdef public dict books
 */

struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange {
  PyObject *(*get_exchange_credentials)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, int __pyx_skip_dispatch);
  PyObject *(*get_book_instance)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, PyObject *, int __pyx_skip_dispatch);
  void (*initialize)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_9exchanges_27abstract_websocket_exchange_25AbstractWebsocketExchange_initialize *__pyx_optional_args);
  int (*_should_authenticate)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, int __pyx_skip_dispatch);
  int (*get_max_handled_pair_with_time_frame)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, int __pyx_skip_dispatch);
  PyObject *(*feed_to_exchange)(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *__pyx_vtabptr_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange;


/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":34
 * 
 * 
 * class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):             # <<<<<<<<<<<<<<
 *     LOGGERS = ["feedhandler"]
 *     CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"
 */

struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector {
  struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange __pyx_base;
  PyObject *(*get_pair_from_exchange)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_exchange_pair)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *, int __pyx_skip_dispatch);
  void (*start)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *);
  void (*subscribe_feeds)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *);
  void (*subscribe_candle_feed)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *);
};
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_vtabptr_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Coroutine.proto */
#define __Pyx_Coroutine_USED
static PyTypeObject *__pyx_CoroutineType = 0;
static PyTypeObject *__pyx_CoroutineAwaitType = 0;
#define __Pyx_Coroutine_CheckExact(obj) (Py_TYPE(obj) == __pyx_CoroutineType)
#define __Pyx_Coroutine_Check(obj) __Pyx_Coroutine_CheckExact(obj)
#define __Pyx_CoroutineAwait_CheckExact(obj) (Py_TYPE(obj) == __pyx_CoroutineAwaitType)
#define __Pyx_Coroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_CoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_Coroutine_init(void);
static PyObject *__Pyx__Coroutine_await(PyObject *coroutine);
typedef struct {
    PyObject_HEAD
    PyObject *coroutine;
} __pyx_CoroutineAwaitObject;
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg);
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args);

/* GetAwaitIter.proto */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o);
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *o);

/* CoroutineYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* IncludeStringH.proto */
#include <string.h>

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* PatchAsyncIO.proto */
static PyObject* __Pyx_patch_asyncio(PyObject* module);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_candle_feed(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_exchange_symbols); /* proto*/
static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_feeds(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto*/
static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_start(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_pair_from_exchange(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_exchange_pair(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'octobot_trading.util.initializable' */
static PyTypeObject *__pyx_ptype_15octobot_trading_4util_13initializable_Initializable = 0;

/* Module declarations from 'octobot_trading.util.config_util' */
static int (*__pyx_f_15octobot_trading_4util_11config_util_is_trader_enabled)(PyObject *, int __pyx_skip_dispatch); /*proto*/
static int (*__pyx_f_15octobot_trading_4util_11config_util_is_trader_simulator_enabled)(PyObject *, int __pyx_skip_dispatch); /*proto*/
static int (*__pyx_f_15octobot_trading_4util_11config_util_is_currency_enabled)(PyObject *, PyObject *, int, int __pyx_skip_dispatch); /*proto*/
static int (*__pyx_f_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_15octobot_trading_4util_11config_util_get_symbols)(PyObject *, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *(*__pyx_f_15octobot_trading_4util_11config_util_get_all_currencies)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_all_currencies *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_15octobot_trading_4util_11config_util_get_pairs)(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_pairs *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_15octobot_trading_4util_11config_util_get_market_pair)(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_market_pair *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_15octobot_trading_4util_11config_util_get_reference_market)(PyObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'octobot_trading.util' */

/* Module declarations from 'octobot_trading.exchanges.abstract_exchange' */
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange = 0;

/* Module declarations from 'octobot_trading.exchanges.exchange_config_data' */
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig = 0;

/* Module declarations from 'octobot_trading.exchanges.exchange_manager' */
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager = 0;

/* Module declarations from 'octobot_trading.exchanges.abstract_websocket_exchange' */
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange = 0;

/* Module declarations from 'octobot_trading.exchanges.connectors.cryptofeed_websocket_connector' */
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info = 0;
static PyTypeObject *__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions = 0;
static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector__set_state(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector"
extern int __pyx_module_is_main_octobot_trading__exchanges__connectors__cryptofeed_websocket_connector;
int __pyx_module_is_main_octobot_trading__exchanges__connectors__cryptofeed_websocket_connector = 0;

/* Implementation of 'octobot_trading.exchanges.connectors.cryptofeed_websocket_connector' */
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_NotImplementedError;
static const char __pyx_k_ID[] = "ID";
static const char __pyx_k_ASK[] = "ASK";
static const char __pyx_k_BID[] = "BID";
static const char __pyx_k_BUY[] = "BUY";
static const char __pyx_k__14[] = "";
static const char __pyx_k__15[] = "*";
static const char __pyx_k__16[] = "-";
static const char __pyx_k_ask[] = "ask";
static const char __pyx_k_bid[] = "bid";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_SELL[] = "SELL";
static const char __pyx_k_SIDE[] = "SIDE";
static const char __pyx_k_SIZE[] = "SIZE";
static const char __pyx_k_TYPE[] = "TYPE";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_asks[] = "asks";
static const char __pyx_k_bids[] = "bids";
static const char __pyx_k_book[] = "book";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_feed[] = "feed";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_pair[] = "pair";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_side[] = "side";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_DEBUG[] = "DEBUG";
static const char __pyx_k_Feeds[] = "Feeds";
static const char __pyx_k_PRICE[] = "PRICE";
static const char __pyx_k_await[] = "__await__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_delta[] = "delta";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_kline[] = "kline";
static const char __pyx_k_price[] = "price";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_trade[] = "trade";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_AMOUNT[] = "AMOUNT";
static const char __pyx_k_CANDLE[] = "CANDLE";
static const char __pyx_k_ECOBIC[] = "ECOBIC";
static const char __pyx_k_SYMBOL[] = "SYMBOL";
static const char __pyx_k_TICKER[] = "TICKER";
static const char __pyx_k_TRADES[] = "TRADES";
static const char __pyx_k_VOLUME[] = "VOLUME";
static const char __pyx_k_amount[] = "amount";
static const char __pyx_k_candle[] = "candle";
static const char __pyx_k_closed[] = "closed";
static const char __pyx_k_config[] = "config";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_symbol[] = "symbol";
static const char __pyx_k_ticker[] = "ticker";
static const char __pyx_k_trades[] = "trades";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_volume[] = "volume";
static const char __pyx_k_CANDLES[] = "CANDLES";
static const char __pyx_k_FUNDING[] = "FUNDING";
static const char __pyx_k_L2_BOOK[] = "L2_BOOK";
static const char __pyx_k_L3_BOOK[] = "L3_BOOK";
static const char __pyx_k_LOGGERS[] = "LOGGERS";
static const char __pyx_k_SIGNALS[] = "SIGNALS";
static const char __pyx_k_asyncio[] = "asyncio";
static const char __pyx_k_funding[] = "funding";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_symbols[] = "symbols";
static const char __pyx_k_add_feed[] = "add_feed";
static const char __pyx_k_channels[] = "channels";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_interval[] = "interval";
static const char __pyx_k_order_id[] = "order_id";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TIMESTAMP[] = "TIMESTAMP";
static const char __pyx_k_ask_price[] = "ask_price";
static const char __pyx_k_bid_price[] = "bid_price";
static const char __pyx_k_book_side[] = "book_side";
static const char __pyx_k_callbacks[] = "callbacks";
static const char __pyx_k_low_price[] = "low_price";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_BOOK_DELTA[] = "BOOK_DELTA";
static const char __pyx_k_TimeFrames[] = "TimeFrames";
static const char __pyx_k_cryptofeed[] = "cryptofeed";
static const char __pyx_k_high_price[] = "high_price";
static const char __pyx_k_leaves_qty[] = "leaves_qty";
static const char __pyx_k_open_price[] = "open_price";
static const char __pyx_k_order_size[] = "order_size";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_time_frame[] = "time_frame";
static const char __pyx_k_FeedHandler[] = "FeedHandler";
static const char __pyx_k_MARKET_INFO[] = "MARKET_INFO";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_UNSUPPORTED[] = "UNSUPPORTED";
static const char __pyx_k_book_prices[] = "book_prices";
static const char __pyx_k_close_price[] = "close_price";
static const char __pyx_k_feedhandler[] = "feedhandler";
static const char __pyx_k_market_info[] = "market_info";
static const char __pyx_k_order_price[] = "order_price";
static const char __pyx_k_symbol_util[] = "symbol_util";
static const char __pyx_k_BookCallback[] = "BookCallback";
static const char __pyx_k_LIQUIDATIONS[] = "LIQUIDATIONS";
static const char __pyx_k_TRANSACTIONS[] = "TRANSACTIONS";
static const char __pyx_k_ask_quantity[] = "ask_quantity";
static const char __pyx_k_bid_quantity[] = "bid_quantity";
static const char __pyx_k_liquidations[] = "liquidations";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_transactions[] = "transactions";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_FUTURES_INDEX[] = "FUTURES_INDEX";
static const char __pyx_k_KLINE_CHANNEL[] = "KLINE_CHANNEL";
static const char __pyx_k_OHLCV_CHANNEL[] = "OHLCV_CHANNEL";
static const char __pyx_k_OPEN_INTEREST[] = "OPEN_INTEREST";
static const char __pyx_k_TradeCallback[] = "TradeCallback";
static const char __pyx_k_asyncio_tasks[] = "asyncio.tasks";
static const char __pyx_k_book_instance[] = "book_instance";
static const char __pyx_k_commons_enums[] = "commons_enums";
static const char __pyx_k_futures_index[] = "futures_index";
static const char __pyx_k_get_feed_name[] = "get_feed_name";
static const char __pyx_k_open_interest[] = "open_interest";
static const char __pyx_k_recent_trades[] = "recent_trades";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_trading_enums[] = "trading_enums";
static const char __pyx_k_CandleCallback[] = "CandleCallback";
static const char __pyx_k_EXCHANGE_FEEDS[] = "EXCHANGE_FEEDS";
static const char __pyx_k_TickerCallback[] = "TickerCallback";
static const char __pyx_k_TradeOrderSide[] = "TradeOrderSide";
static const char __pyx_k_VolumeCallback[] = "VolumeCallback";
static const char __pyx_k_WebsocketFeeds[] = "WebsocketFeeds";
static const char __pyx_k_convert_symbol[] = "convert_symbol";
static const char __pyx_k_new_event_loop[] = "new_event_loop";
static const char __pyx_k_set_event_loop[] = "set_event_loop";
static const char __pyx_k_FundingCallback[] = "FundingCallback";
static const char __pyx_k_candle_interval[] = "candle_interval";
static const char __pyx_k_commons_logging[] = "commons_logging";
static const char __pyx_k_octobot_commons[] = "octobot_commons";
static const char __pyx_k_push_to_channel[] = "push_to_channel";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_MARKET_SEPARATOR[] = "MARKET_SEPARATOR";
static const char __pyx_k_exchange_manager[] = "exchange_manager";
static const char __pyx_k_handle_book_adds[] = "handle_book_adds";
static const char __pyx_k_should_uppercase[] = "should_uppercase";
static const char __pyx_k_symbol_separator[] = "symbol_separator";
static const char __pyx_k_get_exchange_pair[] = "get_exchange_pair";
static const char __pyx_k_receipt_timestamp[] = "receipt_timestamp";
static const char __pyx_k_set_logging_level[] = "set_logging_level";
static const char __pyx_k_trading_constants[] = "trading_constants";
static const char __pyx_k_update_order_book[] = "update_order_book";
static const char __pyx_k_BookUpdateCallback[] = "BookUpdateCallback";
static const char __pyx_k_MarketInfoCallback[] = "MarketInfoCallback";
static const char __pyx_k_ORDER_BOOK_CHANNEL[] = "ORDER_BOOK_CHANNEL";
static const char __pyx_k_abstract_websocket[] = "abstract_websocket";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_candle_closed_only[] = "candle_closed_only";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_cryptofeed_defines[] = "cryptofeed.defines";
static const char __pyx_k_fix_signal_handler[] = "fix_signal_handler";
static const char __pyx_k_LiquidationCallback[] = "LiquidationCallback";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_cryptofeed_callback[] = "cryptofeed.callback";
static const char __pyx_k_FuturesIndexCallback[] = "FuturesIndexCallback";
static const char __pyx_k_OpenInterestCallback[] = "OpenInterestCallback";
static const char __pyx_k_TransactionsCallback[] = "TransactionsCallback";
static const char __pyx_k_cryptofeed_callbacks[] = "cryptofeed_callbacks";
static const char __pyx_k_cryptofeed_constants[] = "cryptofeed_constants";
static const char __pyx_k_new_symbol_separator[] = "new_symbol_separator";
static const char __pyx_k_RECENT_TRADES_CHANNEL[] = "RECENT_TRADES_CHANNEL";
static const char __pyx_k_octobot_commons_enums[] = "octobot_commons.enums";
static const char __pyx_k_octobot_trading_enums[] = "octobot_trading.enums";
static const char __pyx_k_get_pair_from_exchange[] = "get_pair_from_exchange";
static const char __pyx_k_Failed_to_get_market_of[] = "Failed to get market of ";
static const char __pyx_k_octobot_commons_logging[] = "octobot_commons.logging";
static const char __pyx_k_ORDER_BOOK_TICKER_CHANNEL[] = "ORDER_BOOK_TICKER_CHANNEL";
static const char __pyx_k_octobot_trading_constants[] = "octobot_trading.constants";
static const char __pyx_k_octobot_commons_symbol_util[] = "octobot_commons.symbol_util";
static const char __pyx_k_CryptofeedWebsocketConnector[] = "CryptofeedWebsocketConnector";
static const char __pyx_k_ExchangeConstantsOrderColumns[] = "ExchangeConstantsOrderColumns";
static const char __pyx_k_convert_book_prices_to_orders[] = "_convert_book_prices_to_orders";
static const char __pyx_k_get_feed_name_not_implemented[] = "get_feed_name not implemented";
static const char __pyx_k_Failed_to_start_websocket_feed[] = "Failed to start websocket feed : ";
static const char __pyx_k_pyx_unpickle_CryptofeedWebsock[] = "__pyx_unpickle_CryptofeedWebsocketConnector";
static const char __pyx_k_CRYPTOFEED_DEFAULT_MARKET_SEPARA[] = "CRYPTOFEED_DEFAULT_MARKET_SEPARATOR";
static const char __pyx_k_CryptofeedWebsocketConnector___r[] = "CryptofeedWebsocketConnector.__reduce_cython__";
static const char __pyx_k_CryptofeedWebsocketConnector___s[] = "CryptofeedWebsocketConnector.__setstate_cython__";
static const char __pyx_k_CryptofeedWebsocketConnector__co[] = "CryptofeedWebsocketConnector._convert_book_prices_to_orders";
static const char __pyx_k_CryptofeedWebsocketConnector_boo[] = "CryptofeedWebsocketConnector.book";
static const char __pyx_k_CryptofeedWebsocketConnector_can[] = "CryptofeedWebsocketConnector.candle";
static const char __pyx_k_CryptofeedWebsocketConnector_del[] = "CryptofeedWebsocketConnector.delta";
static const char __pyx_k_CryptofeedWebsocketConnector_fix[] = "CryptofeedWebsocketConnector.fix_signal_handler";
static const char __pyx_k_CryptofeedWebsocketConnector_fun[] = "CryptofeedWebsocketConnector.funding";
static const char __pyx_k_CryptofeedWebsocketConnector_fut[] = "CryptofeedWebsocketConnector.futures_index";
static const char __pyx_k_CryptofeedWebsocketConnector_get[] = "CryptofeedWebsocketConnector.get_feed_name";
static const char __pyx_k_CryptofeedWebsocketConnector_liq[] = "CryptofeedWebsocketConnector.liquidations";
static const char __pyx_k_CryptofeedWebsocketConnector_mar[] = "CryptofeedWebsocketConnector.market_info";
static const char __pyx_k_CryptofeedWebsocketConnector_ope[] = "CryptofeedWebsocketConnector.open_interest";
static const char __pyx_k_CryptofeedWebsocketConnector_tic[] = "CryptofeedWebsocketConnector.ticker";
static const char __pyx_k_CryptofeedWebsocketConnector_tra[] = "CryptofeedWebsocketConnector.trade";
static const char __pyx_k_CryptofeedWebsocketConnector_vol[] = "CryptofeedWebsocketConnector.volume";
static const char __pyx_k_ExchangeConstantsOrderBookInfoCo[] = "ExchangeConstantsOrderBookInfoColumns";
static const char __pyx_k_Incompatible_checksums_s_vs_0xe1[] = "Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))";
static const char __pyx_k_octobot_trading_exchanges_abstra[] = "octobot_trading.exchanges.abstract_websocket_exchange";
static const char __pyx_k_octobot_trading_exchanges_connec[] = "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector";
static const char __pyx_k_CryptofeedWebsocketConnector_get_2[] = "CryptofeedWebsocketConnector.get_pair_from_exchange";
static const char __pyx_k_CryptofeedWebsocketConnector_get_3[] = "CryptofeedWebsocketConnector.get_exchange_pair";
static const char __pyx_k_CryptofeedWebsocketConnector_tra_2[] = "CryptofeedWebsocketConnector.transactions";
static const char __pyx_k_octobot_trading_exchanges_connec_2[] = "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py";
static PyObject *__pyx_n_s_AMOUNT;
static PyObject *__pyx_n_s_ASK;
static PyObject *__pyx_n_s_BID;
static PyObject *__pyx_n_s_BOOK_DELTA;
static PyObject *__pyx_n_s_BUY;
static PyObject *__pyx_n_s_BookCallback;
static PyObject *__pyx_n_s_BookUpdateCallback;
static PyObject *__pyx_n_s_CANDLE;
static PyObject *__pyx_n_s_CANDLES;
static PyObject *__pyx_n_s_CRYPTOFEED_DEFAULT_MARKET_SEPARA;
static PyObject *__pyx_n_s_CandleCallback;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector___r;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector___s;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector__co;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_boo;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_can;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_del;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_fix;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_fun;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_fut;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_get;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_get_2;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_get_3;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_liq;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_mar;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_ope;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_tic;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_tra;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_tra_2;
static PyObject *__pyx_n_s_CryptofeedWebsocketConnector_vol;
static PyObject *__pyx_n_s_DEBUG;
static PyObject *__pyx_n_s_ECOBIC;
static PyObject *__pyx_n_s_EXCHANGE_FEEDS;
static PyObject *__pyx_n_s_ExchangeConstantsOrderBookInfoCo;
static PyObject *__pyx_n_s_ExchangeConstantsOrderColumns;
static PyObject *__pyx_n_s_FUNDING;
static PyObject *__pyx_n_s_FUTURES_INDEX;
static PyObject *__pyx_kp_u_Failed_to_get_market_of;
static PyObject *__pyx_kp_u_Failed_to_start_websocket_feed;
static PyObject *__pyx_n_s_FeedHandler;
static PyObject *__pyx_n_s_Feeds;
static PyObject *__pyx_n_s_FundingCallback;
static PyObject *__pyx_n_s_FuturesIndexCallback;
static PyObject *__pyx_n_s_ID;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xe1;
static PyObject *__pyx_n_s_KLINE_CHANNEL;
static PyObject *__pyx_n_s_L2_BOOK;
static PyObject *__pyx_n_s_L3_BOOK;
static PyObject *__pyx_n_s_LIQUIDATIONS;
static PyObject *__pyx_n_s_LOGGERS;
static PyObject *__pyx_n_s_LiquidationCallback;
static PyObject *__pyx_n_s_MARKET_INFO;
static PyObject *__pyx_n_s_MARKET_SEPARATOR;
static PyObject *__pyx_n_s_MarketInfoCallback;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_OHLCV_CHANNEL;
static PyObject *__pyx_n_s_OPEN_INTEREST;
static PyObject *__pyx_n_s_ORDER_BOOK_CHANNEL;
static PyObject *__pyx_n_s_ORDER_BOOK_TICKER_CHANNEL;
static PyObject *__pyx_n_s_OpenInterestCallback;
static PyObject *__pyx_n_s_PRICE;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RECENT_TRADES_CHANNEL;
static PyObject *__pyx_n_s_SELL;
static PyObject *__pyx_n_s_SIDE;
static PyObject *__pyx_n_s_SIGNALS;
static PyObject *__pyx_n_s_SIZE;
static PyObject *__pyx_n_s_SYMBOL;
static PyObject *__pyx_n_s_TICKER;
static PyObject *__pyx_n_s_TIMESTAMP;
static PyObject *__pyx_n_s_TRADES;
static PyObject *__pyx_n_s_TRANSACTIONS;
static PyObject *__pyx_n_s_TYPE;
static PyObject *__pyx_n_s_TickerCallback;
static PyObject *__pyx_n_s_TimeFrames;
static PyObject *__pyx_n_s_TradeCallback;
static PyObject *__pyx_n_s_TradeOrderSide;
static PyObject *__pyx_n_s_TransactionsCallback;
static PyObject *__pyx_n_s_UNSUPPORTED;
static PyObject *__pyx_n_s_VOLUME;
static PyObject *__pyx_n_s_VolumeCallback;
static PyObject *__pyx_n_s_WebsocketFeeds;
static PyObject *__pyx_kp_u__14;
static PyObject *__pyx_n_s__15;
static PyObject *__pyx_kp_u__16;
static PyObject *__pyx_n_s_abstract_websocket;
static PyObject *__pyx_n_s_add_feed;
static PyObject *__pyx_n_s_amount;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_ask;
static PyObject *__pyx_n_s_ask_price;
static PyObject *__pyx_n_s_ask_quantity;
static PyObject *__pyx_n_s_asks;
static PyObject *__pyx_n_s_asyncio;
static PyObject *__pyx_n_s_asyncio_coroutines;
static PyObject *__pyx_n_s_asyncio_tasks;
static PyObject *__pyx_n_s_await;
static PyObject *__pyx_n_s_bid;
static PyObject *__pyx_n_s_bid_price;
static PyObject *__pyx_n_s_bid_quantity;
static PyObject *__pyx_n_s_bids;
static PyObject *__pyx_n_s_book;
static PyObject *__pyx_n_s_book_instance;
static PyObject *__pyx_n_s_book_prices;
static PyObject *__pyx_n_s_book_side;
static PyObject *__pyx_n_s_callbacks;
static PyObject *__pyx_n_s_candle;
static PyObject *__pyx_n_s_candle_closed_only;
static PyObject *__pyx_n_s_candle_interval;
static PyObject *__pyx_n_s_channels;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_close_price;
static PyObject *__pyx_n_s_closed;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_commons_enums;
static PyObject *__pyx_n_s_commons_logging;
static PyObject *__pyx_n_s_config;
static PyObject *__pyx_n_s_convert_book_prices_to_orders;
static PyObject *__pyx_n_s_convert_symbol;
static PyObject *__pyx_n_s_cryptofeed;
static PyObject *__pyx_n_s_cryptofeed_callback;
static PyObject *__pyx_n_s_cryptofeed_callbacks;
static PyObject *__pyx_n_s_cryptofeed_constants;
static PyObject *__pyx_n_s_cryptofeed_defines;
static PyObject *__pyx_n_s_delta;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dict_2;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_exchange_manager;
static PyObject *__pyx_n_s_feed;
static PyObject *__pyx_n_s_feedhandler;
static PyObject *__pyx_n_u_feedhandler;
static PyObject *__pyx_n_s_fix_signal_handler;
static PyObject *__pyx_n_s_funding;
static PyObject *__pyx_n_s_futures_index;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_exchange_pair;
static PyObject *__pyx_n_s_get_feed_name;
static PyObject *__pyx_kp_u_get_feed_name_not_implemented;
static PyObject *__pyx_n_s_get_pair_from_exchange;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_handle_book_adds;
static PyObject *__pyx_n_s_high_price;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_inspect;
static PyObject *__pyx_n_s_interval;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_kline;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_leaves_qty;
static PyObject *__pyx_n_s_liquidations;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_n_s_low_price;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_market_info;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_new_event_loop;
static PyObject *__pyx_n_s_new_symbol_separator;
static PyObject *__pyx_n_s_octobot_commons;
static PyObject *__pyx_n_s_octobot_commons_enums;
static PyObject *__pyx_n_s_octobot_commons_logging;
static PyObject *__pyx_n_s_octobot_commons_symbol_util;
static PyObject *__pyx_n_s_octobot_trading_constants;
static PyObject *__pyx_n_s_octobot_trading_enums;
static PyObject *__pyx_n_s_octobot_trading_exchanges_abstra;
static PyObject *__pyx_n_s_octobot_trading_exchanges_connec;
static PyObject *__pyx_kp_s_octobot_trading_exchanges_connec_2;
static PyObject *__pyx_n_s_open_interest;
static PyObject *__pyx_n_s_open_price;
static PyObject *__pyx_n_s_order_id;
static PyObject *__pyx_n_s_order_price;
static PyObject *__pyx_n_s_order_size;
static PyObject *__pyx_n_s_pair;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_price;
static PyObject *__pyx_n_s_push_to_channel;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_CryptofeedWebsock;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_receipt_timestamp;
static PyObject *__pyx_n_s_recent_trades;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_run;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_event_loop;
static PyObject *__pyx_n_s_set_logging_level;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_should_uppercase;
static PyObject *__pyx_n_s_side;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_symbol;
static PyObject *__pyx_n_s_symbol_separator;
static PyObject *__pyx_n_s_symbol_util;
static PyObject *__pyx_n_s_symbols;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_ticker;
static PyObject *__pyx_n_s_time_frame;
static PyObject *__pyx_n_s_timestamp;
static PyObject *__pyx_n_s_trade;
static PyObject *__pyx_n_s_trades;
static PyObject *__pyx_n_s_trading_constants;
static PyObject *__pyx_n_s_trading_enums;
static PyObject *__pyx_n_s_transactions;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_update_order_book;
static PyObject *__pyx_n_s_use_setstate;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_volume;
static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector___init__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_config, PyObject *__pyx_v_exchange_manager); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_2get_feed_name(CYTHON_UNUSED PyTypeObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_6ticker(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_bid, PyObject *__pyx_v_ask, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_9trade(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_order_id, PyObject *__pyx_v_timestamp, PyObject *__pyx_v_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_12book(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_book, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_15_convert_book_prices_to_orders(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_book_prices, PyObject *__pyx_v_book_side); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_17candle(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_start, CYTHON_UNUSED PyObject *__pyx_v_stop, PyObject *__pyx_v_interval, CYTHON_UNUSED PyObject *__pyx_v_trades, PyObject *__pyx_v_open_price, PyObject *__pyx_v_close_price, PyObject *__pyx_v_high_price, PyObject *__pyx_v_low_price, PyObject *__pyx_v_volume, PyObject *__pyx_v_closed, PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_20delta(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_delta, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_23liquidations(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_side, CYTHON_UNUSED PyObject *__pyx_v_leaves_qty, CYTHON_UNUSED PyObject *__pyx_v_price, CYTHON_UNUSED PyObject *__pyx_v_order_id, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_26funding(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_29open_interest(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_open_interest, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_32volume(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_35futures_index(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_38market_info(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_41transactions(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_44get_pair_from_exchange(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_46get_exchange_pair(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed___get__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto */
static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_2__set__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_4__del__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_48__reduce_cython__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_50__setstate_cython__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_236055009;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_codeobj__2;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__4;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
/* Late includes */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":38
 *     CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"
 * 
 *     def __init__(self, config: object, exchange_manager: object):             # <<<<<<<<<<<<<<
 *         super().__init__(config, exchange_manager)
 *         self.fix_signal_handler()
 */

/* Python wrapper */
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_config = 0;
  PyObject *__pyx_v_exchange_manager = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_config,&__pyx_n_s_exchange_manager,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_config)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exchange_manager)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 38, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 38, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_config = values[0];
    __pyx_v_exchange_manager = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 38, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector___init__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_config, __pyx_v_exchange_manager);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector___init__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_config, PyObject *__pyx_v_exchange_manager) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":39
 * 
 *     def __init__(self, config: object, exchange_manager: object):
 *         super().__init__(config, exchange_manager)             # <<<<<<<<<<<<<<
 *         self.fix_signal_handler()
 *         self.client = cryptofeed.FeedHandler()
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_config, __pyx_v_exchange_manager};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_config, __pyx_v_exchange_manager};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_config);
    __Pyx_GIVEREF(__pyx_v_config);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_config);
    __Pyx_INCREF(__pyx_v_exchange_manager);
    __Pyx_GIVEREF(__pyx_v_exchange_manager);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_exchange_manager);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":40
 *     def __init__(self, config: object, exchange_manager: object):
 *         super().__init__(config, exchange_manager)
 *         self.fix_signal_handler()             # <<<<<<<<<<<<<<
 *         self.client = cryptofeed.FeedHandler()
 *         commons_logging.set_logging_level(self.LOGGERS, logging.DEBUG)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fix_signal_handler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":41
 *         super().__init__(config, exchange_manager)
 *         self.fix_signal_handler()
 *         self.client = cryptofeed.FeedHandler()             # <<<<<<<<<<<<<<
 *         commons_logging.set_logging_level(self.LOGGERS, logging.DEBUG)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FeedHandler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.client);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.client);
  __pyx_v_self->__pyx_base.client = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":42
 *         self.fix_signal_handler()
 *         self.client = cryptofeed.FeedHandler()
 *         commons_logging.set_logging_level(self.LOGGERS, logging.DEBUG)             # <<<<<<<<<<<<<<
 * 
 *         self.callback_by_feed = {
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_commons_logging); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_set_logging_level); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_LOGGERS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DEBUG); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_4, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_4, __pyx_t_6);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":45
 * 
 *         self.callback_by_feed = {
 *             cryptofeed_constants.TRADES: cryptofeed_callbacks.TradeCallback(self.trade),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),
 *             cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),  # pylint: disable=E1101
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TRADES); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TradeCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trade); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":46
 *         self.callback_by_feed = {
 *             cryptofeed_constants.TRADES: cryptofeed_callbacks.TradeCallback(self.trade),
 *             cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),  # pylint: disable=E1101
 *             cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TICKER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_TickerCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ticker); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":47
 *             cryptofeed_constants.TRADES: cryptofeed_callbacks.TradeCallback(self.trade),
 *             cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),
 *             cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),  # pylint: disable=E1101             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.L3_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CANDLES); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_CandleCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_candle); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":48
 *             cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),
 *             cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),  # pylint: disable=E1101
 *             cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.L3_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_L2_BOOK); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_BookCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_book); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":49
 *             cryptofeed_constants.CANDLES: cryptofeed_callbacks.CandleCallback(self.candle),  # pylint: disable=E1101
 *             cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.L3_BOOK: cryptofeed_callbacks.BookCallback(self.book),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),
 *             cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_L3_BOOK); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_BookCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_book); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":50
 *             cryptofeed_constants.L2_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.L3_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),
 *             cryptofeed_constants.BOOK_DELTA: cryptofeed_callbacks.BookUpdateCallback(self.delta),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FUNDING); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FundingCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_funding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":51
 *             cryptofeed_constants.L3_BOOK: cryptofeed_callbacks.BookCallback(self.book),
 *             cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),
 *             cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.BOOK_DELTA: cryptofeed_callbacks.BookUpdateCallback(self.delta),
 *             cryptofeed_constants.VOLUME: cryptofeed_callbacks.VolumeCallback(self.volume),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LIQUIDATIONS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_LiquidationCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_liquidations); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":52
 *             cryptofeed_constants.FUNDING: cryptofeed_callbacks.FundingCallback(self.funding),
 *             cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),
 *             cryptofeed_constants.BOOK_DELTA: cryptofeed_callbacks.BookUpdateCallback(self.delta),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.VOLUME: cryptofeed_callbacks.VolumeCallback(self.volume),
 *             cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BOOK_DELTA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_BookUpdateCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_delta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":53
 *             cryptofeed_constants.LIQUIDATIONS: cryptofeed_callbacks.LiquidationCallback(self.liquidations),
 *             cryptofeed_constants.BOOK_DELTA: cryptofeed_callbacks.BookUpdateCallback(self.delta),
 *             cryptofeed_constants.VOLUME: cryptofeed_callbacks.VolumeCallback(self.volume),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),
 *             cryptofeed_constants.FUTURES_INDEX: cryptofeed_callbacks.FuturesIndexCallback(self.futures_index),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_VOLUME); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_VolumeCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_volume); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":54
 *             cryptofeed_constants.BOOK_DELTA: cryptofeed_callbacks.BookUpdateCallback(self.delta),
 *             cryptofeed_constants.VOLUME: cryptofeed_callbacks.VolumeCallback(self.volume),
 *             cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.FUTURES_INDEX: cryptofeed_callbacks.FuturesIndexCallback(self.futures_index),
 *             cryptofeed_constants.MARKET_INFO: cryptofeed_callbacks.MarketInfoCallback(self.market_info),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_OPEN_INTEREST); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_OpenInterestCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_open_interest); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":55
 *             cryptofeed_constants.VOLUME: cryptofeed_callbacks.VolumeCallback(self.volume),
 *             cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),
 *             cryptofeed_constants.FUTURES_INDEX: cryptofeed_callbacks.FuturesIndexCallback(self.futures_index),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.MARKET_INFO: cryptofeed_callbacks.MarketInfoCallback(self.market_info),
 *             cryptofeed_constants.TRANSACTIONS: cryptofeed_callbacks.TransactionsCallback(self.transactions),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FUTURES_INDEX); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_FuturesIndexCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_futures_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":56
 *             cryptofeed_constants.OPEN_INTEREST: cryptofeed_callbacks.OpenInterestCallback(self.open_interest),
 *             cryptofeed_constants.FUTURES_INDEX: cryptofeed_callbacks.FuturesIndexCallback(self.futures_index),
 *             cryptofeed_constants.MARKET_INFO: cryptofeed_callbacks.MarketInfoCallback(self.market_info),             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.TRANSACTIONS: cryptofeed_callbacks.TransactionsCallback(self.transactions),
 *         }
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MARKET_INFO); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MarketInfoCallback); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":57
 *             cryptofeed_constants.FUTURES_INDEX: cryptofeed_callbacks.FuturesIndexCallback(self.futures_index),
 *             cryptofeed_constants.MARKET_INFO: cryptofeed_callbacks.MarketInfoCallback(self.market_info),
 *             cryptofeed_constants.TRANSACTIONS: cryptofeed_callbacks.TransactionsCallback(self.transactions),             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TRANSACTIONS); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_callbacks); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TransactionsCallback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transactions); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_7, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":44
 *         commons_logging.set_logging_level(self.LOGGERS, logging.DEBUG)
 * 
 *         self.callback_by_feed = {             # <<<<<<<<<<<<<<
 *             cryptofeed_constants.TRADES: cryptofeed_callbacks.TradeCallback(self.trade),
 *             cryptofeed_constants.TICKER: cryptofeed_callbacks.TickerCallback(self.ticker),
 */
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->callback_by_feed);
  __Pyx_DECREF(__pyx_v_self->callback_by_feed);
  __pyx_v_self->callback_by_feed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":38
 *     CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"
 * 
 *     def __init__(self, config: object, exchange_manager: object):             # <<<<<<<<<<<<<<
 *         super().__init__(config, exchange_manager)
 *         self.fix_signal_handler()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":61
 * 
 *     @classmethod
 *     def get_feed_name(cls):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name(PyObject *__pyx_v_cls, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name = {"get_feed_name", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name, METH_NOARGS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name(PyObject *__pyx_v_cls, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_feed_name (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_2get_feed_name(((PyTypeObject*)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_2get_feed_name(CYTHON_UNUSED PyTypeObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_feed_name", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":62
 *     @classmethod
 *     def get_feed_name(cls):
 *         raise NotImplementedError("get_feed_name not implemented")             # <<<<<<<<<<<<<<
 * 
 *     def fix_signal_handler(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 62, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":61
 * 
 *     @classmethod
 *     def get_feed_name(cls):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_feed_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":64
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 *     def fix_signal_handler(self):             # <<<<<<<<<<<<<<
 *         """
 *         Websocket are started in a new thread thus signal handle cannot be used
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler[] = "\n        Websocket are started in a new thread thus signal handle cannot be used\n        add_signal_handler() can only be called from the main thread\n        ";
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler = {"fix_signal_handler", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler, METH_NOARGS, __pyx_doc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fix_signal_handler (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fix_signal_handler", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":69
 *         add_signal_handler() can only be called from the main thread
 *         """
 *         cryptofeed.feedhandler.SIGNALS = ()             # <<<<<<<<<<<<<<
 * 
 *     def subscribe_candle_feed(self, exchange_symbols):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cryptofeed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_feedhandler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_SIGNALS, __pyx_empty_tuple) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":64
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 *     def fix_signal_handler(self):             # <<<<<<<<<<<<<<
 *         """
 *         Websocket are started in a new thread thus signal handle cannot be used
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.fix_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":71
 *         cryptofeed.feedhandler.SIGNALS = ()
 * 
 *     def subscribe_candle_feed(self, exchange_symbols):             # <<<<<<<<<<<<<<
 *         candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]
 * 
 */

static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_candle_feed(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_exchange_symbols) {
  PyObject *__pyx_v_candle_callback = NULL;
  PyObject *__pyx_v_time_frame = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subscribe_candle_feed", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":72
 * 
 *     def subscribe_candle_feed(self, exchange_symbols):
 *         candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]             # <<<<<<<<<<<<<<
 * 
 *         for time_frame in self.time_frames:
 */
  if (unlikely(__pyx_v_self->callback_by_feed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 72, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_EXCHANGE_FEEDS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Feeds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CANDLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->callback_by_feed, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_candle_callback = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":74
 *         candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]
 * 
 *         for time_frame in self.time_frames:             # <<<<<<<<<<<<<<
 *             self.client.add_feed(self.get_feed_name(),
 *                                  candle_interval=time_frame.value,
 */
  if (unlikely(__pyx_v_self->__pyx_base.time_frames == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->__pyx_base.time_frames; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 74, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_time_frame, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":75
 * 
 *         for time_frame in self.time_frames:
 *             self.client.add_feed(self.get_feed_name(),             # <<<<<<<<<<<<<<
 *                                  candle_interval=time_frame.value,
 *                                  candle_closed_only=False,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.client, __pyx_n_s_add_feed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_feed_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":76
 *         for time_frame in self.time_frames:
 *             self.client.add_feed(self.get_feed_name(),
 *                                  candle_interval=time_frame.value,             # <<<<<<<<<<<<<<
 *                                  candle_closed_only=False,
 *                                  symbols=exchange_symbols,
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_time_frame, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_candle_interval, __pyx_t_6) < 0) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":77
 *             self.client.add_feed(self.get_feed_name(),
 *                                  candle_interval=time_frame.value,
 *                                  candle_closed_only=False,             # <<<<<<<<<<<<<<
 *                                  symbols=exchange_symbols,
 *                                  channels=[cryptofeed_constants.CANDLES],  # pylint: disable=E1101
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_candle_closed_only, Py_False) < 0) __PYX_ERR(0, 76, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":78
 *                                  candle_interval=time_frame.value,
 *                                  candle_closed_only=False,
 *                                  symbols=exchange_symbols,             # <<<<<<<<<<<<<<
 *                                  channels=[cryptofeed_constants.CANDLES],  # pylint: disable=E1101
 *                                  callbacks={cryptofeed_constants.CANDLES: candle_callback})  # pylint: disable=E1101
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_symbols, __pyx_v_exchange_symbols) < 0) __PYX_ERR(0, 76, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":79
 *                                  candle_closed_only=False,
 *                                  symbols=exchange_symbols,
 *                                  channels=[cryptofeed_constants.CANDLES],  # pylint: disable=E1101             # <<<<<<<<<<<<<<
 *                                  callbacks={cryptofeed_constants.CANDLES: candle_callback})  # pylint: disable=E1101
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_CANDLES); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_channels, __pyx_t_6) < 0) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":80
 *                                  symbols=exchange_symbols,
 *                                  channels=[cryptofeed_constants.CANDLES],  # pylint: disable=E1101
 *                                  callbacks={cryptofeed_constants.CANDLES: candle_callback})  # pylint: disable=E1101             # <<<<<<<<<<<<<<
 * 
 *     def subscribe_feeds(self):
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_CANDLES); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_6, __pyx_t_8, __pyx_v_candle_callback) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_callbacks, __pyx_t_6) < 0) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":75
 * 
 *         for time_frame in self.time_frames:
 *             self.client.add_feed(self.get_feed_name(),             # <<<<<<<<<<<<<<
 *                                  candle_interval=time_frame.value,
 *                                  candle_closed_only=False,
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":74
 *         candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]
 * 
 *         for time_frame in self.time_frames:             # <<<<<<<<<<<<<<
 *             self.client.add_feed(self.get_feed_name(),
 *                                  candle_interval=time_frame.value,
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":71
 *         cryptofeed.feedhandler.SIGNALS = ()
 * 
 *     def subscribe_candle_feed(self, exchange_symbols):             # <<<<<<<<<<<<<<
 *         candle_callback = self.callback_by_feed[self.EXCHANGE_FEEDS[Feeds.CANDLE]]
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.subscribe_candle_feed", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candle_callback);
  __Pyx_XDECREF(__pyx_v_time_frame);
  __Pyx_RefNannyFinishContext();
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":82
 *                                  callbacks={cryptofeed_constants.CANDLES: candle_callback})  # pylint: disable=E1101
 * 
 *     def subscribe_feeds(self):             # <<<<<<<<<<<<<<
 *         exchange_symbols = [
 *             self.get_exchange_pair(pair)
 */

static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_feeds(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  PyObject *__pyx_v_exchange_symbols = NULL;
  PyObject *__pyx_v_callbacks = NULL;
  PyObject *__pyx_7genexpr__pyx_v_pair = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_channel = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_channel = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subscribe_feeds", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":83
 * 
 *     def subscribe_feeds(self):
 *         exchange_symbols = [             # <<<<<<<<<<<<<<
 *             self.get_exchange_pair(pair)
 *             for pair in self.pairs
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":85
 *         exchange_symbols = [
 *             self.get_exchange_pair(pair)
 *             for pair in self.pairs             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
    if (unlikely(__pyx_v_self->__pyx_base.pairs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 85, __pyx_L5_error)
    }
    __pyx_t_2 = __pyx_v_self->__pyx_base.pairs; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 85, __pyx_L5_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 85, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_pair, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":84
 *     def subscribe_feeds(self):
 *         exchange_symbols = [
 *             self.get_exchange_pair(pair)             # <<<<<<<<<<<<<<
 *             for pair in self.pairs
 *         ]
 */
      if (!(likely(PyUnicode_CheckExact(__pyx_7genexpr__pyx_v_pair))||((__pyx_7genexpr__pyx_v_pair) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_7genexpr__pyx_v_pair)->tp_name), 0))) __PYX_ERR(0, 84, __pyx_L5_error)
      __pyx_t_4 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self->__pyx_base.__pyx_vtab)->get_exchange_pair(__pyx_v_self, ((PyObject*)__pyx_7genexpr__pyx_v_pair), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 83, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":85
 *         exchange_symbols = [
 *             self.get_exchange_pair(pair)
 *             for pair in self.pairs             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_pair); __pyx_7genexpr__pyx_v_pair = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_pair); __pyx_7genexpr__pyx_v_pair = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_exchange_symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":88
 *         ]
 * 
 *         if self.EXCHANGE_FEEDS.get(Feeds.CANDLE):             # <<<<<<<<<<<<<<
 *             self.subscribe_candle_feed(exchange_symbols)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_EXCHANGE_FEEDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Feeds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CANDLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":89
 * 
 *         if self.EXCHANGE_FEEDS.get(Feeds.CANDLE):
 *             self.subscribe_candle_feed(exchange_symbols)             # <<<<<<<<<<<<<<
 * 
 *         # drop unsupported channels
 */
    ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self->__pyx_base.__pyx_vtab)->subscribe_candle_feed(__pyx_v_self, __pyx_v_exchange_symbols);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":88
 *         ]
 * 
 *         if self.EXCHANGE_FEEDS.get(Feeds.CANDLE):             # <<<<<<<<<<<<<<
 *             self.subscribe_candle_feed(exchange_symbols)
 * 
 */
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":92
 * 
 *         # drop unsupported channels
 *         self.channels = [channel for channel in self.channels if channel not in [Feeds.UNSUPPORTED,             # <<<<<<<<<<<<<<
 *                                                                                  self.EXCHANGE_FEEDS.get(
 *                                                                                      Feeds.CANDLE)]]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_self->__pyx_base.channels == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 92, __pyx_L12_error)
    }
    __pyx_t_4 = __pyx_v_self->__pyx_base.channels; __Pyx_INCREF(__pyx_t_4); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 92, __pyx_L12_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_channel, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":93
 *         # drop unsupported channels
 *         self.channels = [channel for channel in self.channels if channel not in [Feeds.UNSUPPORTED,
 *                                                                                  self.EXCHANGE_FEEDS.get(             # <<<<<<<<<<<<<<
 *                                                                                      Feeds.CANDLE)]]
 *         callbacks = {
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_EXCHANGE_FEEDS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 93, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":94
 *         self.channels = [channel for channel in self.channels if channel not in [Feeds.UNSUPPORTED,
 *                                                                                  self.EXCHANGE_FEEDS.get(
 *                                                                                      Feeds.CANDLE)]]             # <<<<<<<<<<<<<<
 *         callbacks = {
 *             channel: self.callback_by_feed[channel]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Feeds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CANDLE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 94, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 93, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":92
 * 
 *         # drop unsupported channels
 *         self.channels = [channel for channel in self.channels if channel not in [Feeds.UNSUPPORTED,             # <<<<<<<<<<<<<<
 *                                                                                  self.EXCHANGE_FEEDS.get(
 *                                                                                      Feeds.CANDLE)]]
 */
      __Pyx_INCREF(__pyx_8genexpr1__pyx_v_channel);
      __pyx_t_7 = __pyx_8genexpr1__pyx_v_channel;
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Feeds); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_UNSUPPORTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_7, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_9) {
      } else {
        __pyx_t_6 = __pyx_t_9;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_7, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 92, __pyx_L12_error)
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 92, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_6 = __pyx_t_9;
      __pyx_L16_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_9 = (__pyx_t_6 != 0);
      if (__pyx_t_9) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr1__pyx_v_channel))) __PYX_ERR(0, 92, __pyx_L12_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_channel); __pyx_8genexpr1__pyx_v_channel = 0;
    goto __pyx_L18_exit_scope;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_channel); __pyx_8genexpr1__pyx_v_channel = 0;
    goto __pyx_L1_error;
    __pyx_L18_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.channels);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.channels);
  __pyx_v_self->__pyx_base.channels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":95
 *                                                                                  self.EXCHANGE_FEEDS.get(
 *                                                                                      Feeds.CANDLE)]]
 *         callbacks = {             # <<<<<<<<<<<<<<
 *             channel: self.callback_by_feed[channel]
 *             for channel in self.channels
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L21_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":97
 *         callbacks = {
 *             channel: self.callback_by_feed[channel]
 *             for channel in self.channels             # <<<<<<<<<<<<<<
 *         }
 *         self.client.add_feed(self.get_feed_name(),
 */
    if (unlikely(__pyx_v_self->__pyx_base.channels == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 97, __pyx_L21_error)
    }
    __pyx_t_4 = __pyx_v_self->__pyx_base.channels; __Pyx_INCREF(__pyx_t_4); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 97, __pyx_L21_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 97, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_channel, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":96
 *                                                                                      Feeds.CANDLE)]]
 *         callbacks = {
 *             channel: self.callback_by_feed[channel]             # <<<<<<<<<<<<<<
 *             for channel in self.channels
 *         }
 */
      if (unlikely(__pyx_v_self->callback_by_feed == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 96, __pyx_L21_error)
      }
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_self->callback_by_feed, __pyx_8genexpr2__pyx_v_channel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_channel, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 96, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":97
 *         callbacks = {
 *             channel: self.callback_by_feed[channel]
 *             for channel in self.channels             # <<<<<<<<<<<<<<
 *         }
 *         self.client.add_feed(self.get_feed_name(),
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_channel); __pyx_8genexpr2__pyx_v_channel = 0;
    goto __pyx_L24_exit_scope;
    __pyx_L21_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_channel); __pyx_8genexpr2__pyx_v_channel = 0;
    goto __pyx_L1_error;
    __pyx_L24_exit_scope:;
  } /* exit inner scope */
  __pyx_v_callbacks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":99
 *             for channel in self.channels
 *         }
 *         self.client.add_feed(self.get_feed_name(),             # <<<<<<<<<<<<<<
 *                              symbols=exchange_symbols,
 *                              channels=self.channels,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.client, __pyx_n_s_add_feed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_feed_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":100
 *         }
 *         self.client.add_feed(self.get_feed_name(),
 *                              symbols=exchange_symbols,             # <<<<<<<<<<<<<<
 *                              channels=self.channels,
 *                              callbacks=callbacks)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_symbols, __pyx_v_exchange_symbols) < 0) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":101
 *         self.client.add_feed(self.get_feed_name(),
 *                              symbols=exchange_symbols,
 *                              channels=self.channels,             # <<<<<<<<<<<<<<
 *                              callbacks=callbacks)
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_channels, __pyx_v_self->__pyx_base.channels) < 0) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":102
 *                              symbols=exchange_symbols,
 *                              channels=self.channels,
 *                              callbacks=callbacks)             # <<<<<<<<<<<<<<
 * 
 *     def start(self):
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_callbacks, __pyx_v_callbacks) < 0) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":99
 *             for channel in self.channels
 *         }
 *         self.client.add_feed(self.get_feed_name(),             # <<<<<<<<<<<<<<
 *                              symbols=exchange_symbols,
 *                              channels=self.channels,
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":82
 *                                  callbacks={cryptofeed_constants.CANDLES: candle_callback})  # pylint: disable=E1101
 * 
 *     def subscribe_feeds(self):             # <<<<<<<<<<<<<<
 *         exchange_symbols = [
 *             self.get_exchange_pair(pair)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.subscribe_feeds", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exchange_symbols);
  __Pyx_XDECREF(__pyx_v_callbacks);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_pair);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_channel);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_channel);
  __Pyx_RefNannyFinishContext();
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":104
 *                              callbacks=callbacks)
 * 
 *     def start(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.subscribe_feeds()
 */

static void __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_start(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_loop = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":105
 * 
 *     def start(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.subscribe_feeds()
 *         except Exception as e:
 */
  {
    (void)__pyx_t_1; (void)__pyx_t_2; (void)__pyx_t_3; /* mark used */
    /*try:*/ {

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":106
 *     def start(self):
 *         try:
 *             self.subscribe_feeds()             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.logger.error(f"Failed to subscribe when creating websocket feed : {e}")
 */
      ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self->__pyx_base.__pyx_vtab)->subscribe_feeds(__pyx_v_self);

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":105
 * 
 *     def start(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.subscribe_feeds()
 *         except Exception as e:
 */
    }
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":109
 *         except Exception as e:
 *             self.logger.error(f"Failed to subscribe when creating websocket feed : {e}")
 *         try:             # <<<<<<<<<<<<<<
 *             # without this two lines `There is no current event loop in thread`
 *             # is raised when calling `client.run()`
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":112
 *             # without this two lines `There is no current event loop in thread`
 *             # is raised when calling `client.run()`
 *             loop = asyncio.new_event_loop()             # <<<<<<<<<<<<<<
 *             asyncio.set_event_loop(loop)
 *             self.client.run()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 112, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_new_event_loop); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 112, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_loop = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":113
 *             # is raised when calling `client.run()`
 *             loop = asyncio.new_event_loop()
 *             asyncio.set_event_loop(loop)             # <<<<<<<<<<<<<<
 *             self.client.run()
 *         except Exception as e:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 113, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_set_event_loop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_loop) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_loop);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 113, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":114
 *             loop = asyncio.new_event_loop()
 *             asyncio.set_event_loop(loop)
 *             self.client.run()             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.client, __pyx_n_s_run); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 114, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 114, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":109
 *         except Exception as e:
 *             self.logger.error(f"Failed to subscribe when creating websocket feed : {e}")
 *         try:             # <<<<<<<<<<<<<<
 *             # without this two lines `There is no current event loop in thread`
 *             # is raised when calling `client.run()`
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":115
 *             asyncio.set_event_loop(loop)
 *             self.client.run()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 115, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":116
 *             self.client.run()
 *         except Exception as e:
 *             self.logger.error(f"Failed to start websocket feed : {e}")             # <<<<<<<<<<<<<<
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.logger, __pyx_n_s_error); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 116, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_v_e, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 116, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Failed_to_start_websocket_feed, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 116, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_11);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 116, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":115
 *             asyncio.set_event_loop(loop)
 *             self.client.run()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          goto __pyx_L21;
        }
        __pyx_L20_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_7 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L11_except_error;
        }
        __pyx_L21:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L10_exception_handled;
    }
    goto __pyx_L11_except_error;
    __pyx_L11_except_error:;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":109
 *         except Exception as e:
 *             self.logger.error(f"Failed to subscribe when creating websocket feed : {e}")
 *         try:             # <<<<<<<<<<<<<<
 *             # without this two lines `There is no current event loop in thread`
 *             # is raised when calling `client.run()`
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L10_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L14_try_end:;
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":104
 *                              callbacks=callbacks)
 * 
 *     def start(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.subscribe_feeds()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.start", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_loop);
  __Pyx_RefNannyFinishContext();
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_8generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":118
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker = {"ticker", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  PyObject *__pyx_v_symbol = 0;
  PyObject *__pyx_v_bid = 0;
  PyObject *__pyx_v_ask = 0;
  CYTHON_UNUSED PyObject *__pyx_v_timestamp = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ticker (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_bid,&__pyx_n_s_ask,&__pyx_n_s_timestamp,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, 1); __PYX_ERR(0, 118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, 2); __PYX_ERR(0, 118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, 3); __PYX_ERR(0, 118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, 4); __PYX_ERR(0, 118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, 5); __PYX_ERR(0, 118, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ticker") < 0)) __PYX_ERR(0, 118, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_bid = values[2];
    __pyx_v_ask = values[3];
    __pyx_v_timestamp = values[4];
    __pyx_v_receipt_timestamp = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ticker", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 118, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.ticker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_6ticker(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_bid, __pyx_v_ask, __pyx_v_timestamp, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_6ticker(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_bid, PyObject *__pyx_v_ask, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ticker", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 118, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_bid = __pyx_v_bid;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bid);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bid);
  __pyx_cur_scope->__pyx_v_ask = __pyx_v_ask;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ask);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ask);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_8generator, __pyx_codeobj__2, (PyObject *) __pyx_cur_scope, __pyx_n_s_ticker, __pyx_n_s_CryptofeedWebsocketConnector_tic, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.ticker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_8generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ticker", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 118, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":119
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_symbol); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 119, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":120
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)             # <<<<<<<<<<<<<<
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,
 *                                        symbol=symbol,
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_symbol))||((__pyx_cur_scope->__pyx_v_symbol) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_cur_scope->__pyx_v_symbol)->tp_name), 0))) __PYX_ERR(0, 120, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->get_pair_from_exchange(__pyx_cur_scope->__pyx_v_self, ((PyObject*)__pyx_cur_scope->__pyx_v_symbol), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_symbol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_symbol, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":121
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        ask_quantity=ask,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_push_to_channel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_trading_constants); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ORDER_BOOK_TICKER_CHANNEL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":122
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,
 *                                        symbol=symbol,             # <<<<<<<<<<<<<<
 *                                        ask_quantity=ask,
 *                                        ask_price=None,
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_symbol, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":123
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,
 *                                        symbol=symbol,
 *                                        ask_quantity=ask,             # <<<<<<<<<<<<<<
 *                                        ask_price=None,
 *                                        bid_quantity=bid,
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ask_quantity, __pyx_cur_scope->__pyx_v_ask) < 0) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":124
 *                                        symbol=symbol,
 *                                        ask_quantity=ask,
 *                                        ask_price=None,             # <<<<<<<<<<<<<<
 *                                        bid_quantity=bid,
 *                                        bid_price=None)
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ask_price, Py_None) < 0) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":125
 *                                        ask_quantity=ask,
 *                                        ask_price=None,
 *                                        bid_quantity=bid,             # <<<<<<<<<<<<<<
 *                                        bid_price=None)
 * 
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_bid_quantity, __pyx_cur_scope->__pyx_v_bid) < 0) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":126
 *                                        ask_price=None,
 *                                        bid_quantity=bid,
 *                                        bid_price=None)             # <<<<<<<<<<<<<<
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_bid_price, Py_None) < 0) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":121
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        ask_quantity=ask,
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L5_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 121, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_Occurred();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(0, 121, __pyx_L1_error)
      }
    }

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":119
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_TICKER_CHANNEL,
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":118
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("ticker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_11generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":128
 *                                        bid_price=None)
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade = {"trade", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  PyObject *__pyx_v_symbol = 0;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_timestamp = 0;
  PyObject *__pyx_v_side = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_price = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trade (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_order_id,&__pyx_n_s_timestamp,&__pyx_n_s_side,&__pyx_n_s_amount,&__pyx_n_s_price,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 1); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 2); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 3); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 4); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_amount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 5); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 6); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, 7); __PYX_ERR(0, 128, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "trade") < 0)) __PYX_ERR(0, 128, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_order_id = values[2];
    __pyx_v_timestamp = values[3];
    __pyx_v_side = values[4];
    __pyx_v_amount = values[5];
    __pyx_v_price = values[6];
    __pyx_v_receipt_timestamp = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("trade", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 128, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.trade", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_9trade(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_order_id, __pyx_v_timestamp, __pyx_v_side, __pyx_v_amount, __pyx_v_price, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_9trade(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_order_id, PyObject *__pyx_v_timestamp, PyObject *__pyx_v_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trade", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 128, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_order_id = __pyx_v_order_id;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_order_id);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_order_id);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_side = __pyx_v_side;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_side);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_side);
  __pyx_cur_scope->__pyx_v_amount = __pyx_v_amount;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_amount);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_amount);
  __pyx_cur_scope->__pyx_v_price = __pyx_v_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_price);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_11generator1, __pyx_codeobj__3, (PyObject *) __pyx_cur_scope, __pyx_n_s_trade, __pyx_n_s_CryptofeedWebsocketConnector_tra, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.trade", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_11generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trade", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":129
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_symbol); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 129, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":130
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)             # <<<<<<<<<<<<<<
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
 *                                        symbol=symbol,
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_symbol))||((__pyx_cur_scope->__pyx_v_symbol) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_cur_scope->__pyx_v_symbol)->tp_name), 0))) __PYX_ERR(0, 130, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->get_pair_from_exchange(__pyx_cur_scope->__pyx_v_self, ((PyObject*)__pyx_cur_scope->__pyx_v_symbol), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_symbol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_symbol, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":131
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        recent_trades=[{
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_push_to_channel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_trading_constants); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_RECENT_TRADES_CHANNEL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":132
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
 *                                        symbol=symbol,             # <<<<<<<<<<<<<<
 *                                        recent_trades=[{
 *                                            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_symbol, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 132, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":134
 *                                        symbol=symbol,
 *                                        recent_trades=[{
 *                                            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
 *                                            trading_enums.ExchangeConstantsOrderColumns.ID.value: order_id,
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_TIMESTAMP); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_7, __pyx_cur_scope->__pyx_v_timestamp) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":135
 *                                        recent_trades=[{
 *                                            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.ID.value: order_id,
 *                                            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_SYMBOL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_6, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":136
 *                                            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
 *                                            trading_enums.ExchangeConstantsOrderColumns.ID.value: order_id,             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: side,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_7, __pyx_cur_scope->__pyx_v_order_id) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":137
 *                                            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
 *                                            trading_enums.ExchangeConstantsOrderColumns.ID.value: order_id,
 *                                            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: side,
 *                                            trading_enums.ExchangeConstantsOrderColumns.PRICE.value: float(price),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TYPE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_6, Py_None) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":138
 *                                            trading_enums.ExchangeConstantsOrderColumns.ID.value: order_id,
 *                                            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: side,             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.PRICE.value: float(price),
 *                                            trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value: float(amount)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_SIDE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_7, __pyx_cur_scope->__pyx_v_side) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":139
 *                                            trading_enums.ExchangeConstantsOrderColumns.TYPE.value: None,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: side,
 *                                            trading_enums.ExchangeConstantsOrderColumns.PRICE.value: float(price),             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value: float(amount)
 *                                        }])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PRICE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_price); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_6, __pyx_t_7) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":140
 *                                            trading_enums.ExchangeConstantsOrderColumns.SIDE.value: side,
 *                                            trading_enums.ExchangeConstantsOrderColumns.PRICE.value: float(price),
 *                                            trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value: float(amount)             # <<<<<<<<<<<<<<
 *                                        }])
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ExchangeConstantsOrderColumns); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_AMOUNT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_amount); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_5, __pyx_t_6, __pyx_t_7) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":133
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
 *                                        symbol=symbol,
 *                                        recent_trades=[{             # <<<<<<<<<<<<<<
 *                                            trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,
 *                                            trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
 */
    __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_recent_trades, __pyx_t_7) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":131
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        recent_trades=[{
 */
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L5_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 131, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_Occurred();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(0, 131, __pyx_L1_error)
      }
    }

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":129
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             await self.push_to_channel(trading_constants.RECENT_TRADES_CHANNEL,
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":128
 *                                        bid_price=None)
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("trade", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_14generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":143
 *                                        }])
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book = {"book", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  PyObject *__pyx_v_symbol = 0;
  PyObject *__pyx_v_book = 0;
  CYTHON_UNUSED PyObject *__pyx_v_timestamp = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("book (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_book,&__pyx_n_s_timestamp,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("book", 1, 5, 5, 1); __PYX_ERR(0, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_book)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("book", 1, 5, 5, 2); __PYX_ERR(0, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("book", 1, 5, 5, 3); __PYX_ERR(0, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("book", 1, 5, 5, 4); __PYX_ERR(0, 143, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "book") < 0)) __PYX_ERR(0, 143, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_book = values[2];
    __pyx_v_timestamp = values[3];
    __pyx_v_receipt_timestamp = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("book", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 143, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_12book(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_book, __pyx_v_timestamp, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_12book(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, PyObject *__pyx_v_book, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("book", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 143, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_book = __pyx_v_book;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_book);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_book);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_14generator2, __pyx_codeobj__4, (PyObject *) __pyx_cur_scope, __pyx_n_s_book, __pyx_n_s_CryptofeedWebsocketConnector_boo, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_14generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("book", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 143, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":144
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             book_instance = self.get_book_instance(symbol)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_symbol); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 144, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":145
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)             # <<<<<<<<<<<<<<
 *             book_instance = self.get_book_instance(symbol)
 * 
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_symbol))||((__pyx_cur_scope->__pyx_v_symbol) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_cur_scope->__pyx_v_symbol)->tp_name), 0))) __PYX_ERR(0, 145, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->get_pair_from_exchange(__pyx_cur_scope->__pyx_v_self, ((PyObject*)__pyx_cur_scope->__pyx_v_symbol), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_symbol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_symbol, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":146
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             book_instance = self.get_book_instance(symbol)             # <<<<<<<<<<<<<<
 * 
 *             book_instance.handle_book_adds(
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_symbol))||((__pyx_cur_scope->__pyx_v_symbol) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_cur_scope->__pyx_v_symbol)->tp_name), 0))) __PYX_ERR(0, 146, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_book_instance(((struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange *)__pyx_cur_scope->__pyx_v_self), ((PyObject*)__pyx_cur_scope->__pyx_v_symbol), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v_book_instance = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":148
 *             book_instance = self.get_book_instance(symbol)
 * 
 *             book_instance.handle_book_adds(             # <<<<<<<<<<<<<<
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.ASK],
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_book_instance, __pyx_n_s_handle_book_adds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":149
 * 
 *             book_instance.handle_book_adds(
 *                 self._convert_book_prices_to_orders(             # <<<<<<<<<<<<<<
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_convert_book_prices_to_orders); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":150
 *             book_instance.handle_book_adds(
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.ASK],             # <<<<<<<<<<<<<<
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 *                 self._convert_book_prices_to_orders(
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ASK); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_book, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_book_prices, __pyx_t_6) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":151
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +             # <<<<<<<<<<<<<<
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.BID],
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TradeOrderSide); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_SELL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_book_side, __pyx_t_7) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":149
 * 
 *             book_instance.handle_book_adds(
 *                 self._convert_book_prices_to_orders(             # <<<<<<<<<<<<<<
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 */
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":152
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 *                 self._convert_book_prices_to_orders(             # <<<<<<<<<<<<<<
 *                     book_prices=book[cryptofeed_constants.BID],
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_convert_book_prices_to_orders); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":153
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.BID],             # <<<<<<<<<<<<<<
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))
 * 
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_cryptofeed_constants); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_BID); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_book, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_book_prices, __pyx_t_6) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":154
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.BID],
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))             # <<<<<<<<<<<<<<
 * 
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_enums); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_TradeOrderSide); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_BUY); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_book_side, __pyx_t_8) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":152
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +
 *                 self._convert_book_prices_to_orders(             # <<<<<<<<<<<<<<
 *                     book_prices=book[cryptofeed_constants.BID],
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))
 */
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":151
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.ASK],
 *                     book_side=trading_enums.TradeOrderSide.SELL.value) +             # <<<<<<<<<<<<<<
 *                 self._convert_book_prices_to_orders(
 *                     book_prices=book[cryptofeed_constants.BID],
 */
    __pyx_t_4 = PyNumber_Add(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":156
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))
 * 
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        asks=book_instance.asks,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_push_to_channel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_trading_constants); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ORDER_BOOK_CHANNEL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":157
 * 
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,
 *                                        symbol=symbol,             # <<<<<<<<<<<<<<
 *                                        asks=book_instance.asks,
 *                                        bids=book_instance.bids,
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_symbol, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 157, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":158
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,
 *                                        symbol=symbol,
 *                                        asks=book_instance.asks,             # <<<<<<<<<<<<<<
 *                                        bids=book_instance.bids,
 *                                        update_order_book=False)
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_book_instance, __pyx_n_s_asks); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_asks, __pyx_t_8) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":159
 *                                        symbol=symbol,
 *                                        asks=book_instance.asks,
 *                                        bids=book_instance.bids,             # <<<<<<<<<<<<<<
 *                                        update_order_book=False)
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_book_instance, __pyx_n_s_bids); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_bids, __pyx_t_8) < 0) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":160
 *                                        asks=book_instance.asks,
 *                                        bids=book_instance.bids,
 *                                        update_order_book=False)             # <<<<<<<<<<<<<<
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_update_order_book, Py_False) < 0) __PYX_ERR(0, 157, __pyx_L1_error)

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":156
 *                     book_side=trading_enums.TradeOrderSide.BUY.value))
 * 
 *             await self.push_to_channel(trading_constants.ORDER_BOOK_CHANNEL,             # <<<<<<<<<<<<<<
 *                                        symbol=symbol,
 *                                        asks=book_instance.asks,
 */
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L5_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 156, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_Occurred();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(0, 156, __pyx_L1_error)
      }
    }

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":144
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             book_instance = self.get_book_instance(symbol)
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":143
 *                                        }])
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":162
 *                                        update_order_book=False)
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):             # <<<<<<<<<<<<<<
 *         return [
 *             {
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders = {"_convert_book_prices_to_orders", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_book_prices = 0;
  PyObject *__pyx_v_book_side = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_book_prices_to_orders (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_book_prices,&__pyx_n_s_book_side,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_book_prices)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_book_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_book_prices_to_orders", 1, 2, 2, 1); __PYX_ERR(0, 162, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_convert_book_prices_to_orders") < 0)) __PYX_ERR(0, 162, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_book_prices = values[0];
    __pyx_v_book_side = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_book_prices_to_orders", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 162, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector._convert_book_prices_to_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_15_convert_book_prices_to_orders(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_book_prices, __pyx_v_book_side);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_15_convert_book_prices_to_orders(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_book_prices, PyObject *__pyx_v_book_side) {
  PyObject *__pyx_8genexpr3__pyx_v_order_price = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_order_size = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_book_prices_to_orders", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":163
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):
 *         return [             # <<<<<<<<<<<<<<
 *             {
 *                 ECOBIC.PRICE.value: float(order_price),
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":169
 *                 ECOBIC.SIDE.value: book_side,
 *             }
 *             for order_price, order_size in book_prices.items()             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_book_prices == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 169, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_book_prices, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 169, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 169, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_order_price, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_order_size, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":165
 *         return [
 *             {
 *                 ECOBIC.PRICE.value: float(order_price),             # <<<<<<<<<<<<<<
 *                 ECOBIC.SIZE.value: float(order_size),
 *                 ECOBIC.SIDE.value: book_side,
 */
      __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ECOBIC); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PRICE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyNumber_Float(__pyx_8genexpr3__pyx_v_order_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_t_9) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":166
 *             {
 *                 ECOBIC.PRICE.value: float(order_price),
 *                 ECOBIC.SIZE.value: float(order_size),             # <<<<<<<<<<<<<<
 *                 ECOBIC.SIDE.value: book_side,
 *             }
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ECOBIC); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 166, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 166, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyNumber_Float(__pyx_8genexpr3__pyx_v_order_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_7, __pyx_t_9, __pyx_t_6) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":167
 *                 ECOBIC.PRICE.value: float(order_price),
 *                 ECOBIC.SIZE.value: float(order_size),
 *                 ECOBIC.SIDE.value: book_side,             # <<<<<<<<<<<<<<
 *             }
 *             for order_price, order_size in book_prices.items()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ECOBIC); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_SIDE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_v_book_side) < 0) __PYX_ERR(0, 165, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 163, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_price); __pyx_8genexpr3__pyx_v_order_price = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_size); __pyx_8genexpr3__pyx_v_order_size = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_price); __pyx_8genexpr3__pyx_v_order_price = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_size); __pyx_8genexpr3__pyx_v_order_size = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":162
 *                                        update_order_book=False)
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):             # <<<<<<<<<<<<<<
 *         return [
 *             {
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector._convert_book_prices_to_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_price);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_order_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_19generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":172
 *         ]
 * 
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,             # <<<<<<<<<<<<<<
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle = {"candle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  PyObject *__pyx_v_symbol = 0;
  CYTHON_UNUSED PyObject *__pyx_v_start = 0;
  CYTHON_UNUSED PyObject *__pyx_v_stop = 0;
  PyObject *__pyx_v_interval = 0;
  CYTHON_UNUSED PyObject *__pyx_v_trades = 0;
  PyObject *__pyx_v_open_price = 0;
  PyObject *__pyx_v_close_price = 0;
  PyObject *__pyx_v_high_price = 0;
  PyObject *__pyx_v_low_price = 0;
  PyObject *__pyx_v_volume = 0;
  PyObject *__pyx_v_closed = 0;
  PyObject *__pyx_v_timestamp = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("candle (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_start,&__pyx_n_s_stop,&__pyx_n_s_interval,&__pyx_n_s_trades,&__pyx_n_s_open_price,&__pyx_n_s_close_price,&__pyx_n_s_high_price,&__pyx_n_s_low_price,&__pyx_n_s_volume,&__pyx_n_s_closed,&__pyx_n_s_timestamp,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 1); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 2); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 3); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interval)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 4); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trades)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 5); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_open_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 6); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_close_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 7); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_high_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 8); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_low_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 9); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_volume)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 10); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_closed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 11); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 12); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, 13); __PYX_ERR(0, 172, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "candle") < 0)) __PYX_ERR(0, 172, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_start = values[2];
    __pyx_v_stop = values[3];
    __pyx_v_interval = values[4];
    __pyx_v_trades = values[5];
    __pyx_v_open_price = values[6];
    __pyx_v_close_price = values[7];
    __pyx_v_high_price = values[8];
    __pyx_v_low_price = values[9];
    __pyx_v_volume = values[10];
    __pyx_v_closed = values[11];
    __pyx_v_timestamp = values[12];
    __pyx_v_receipt_timestamp = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("candle", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.candle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_17candle(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_start, __pyx_v_stop, __pyx_v_interval, __pyx_v_trades, __pyx_v_open_price, __pyx_v_close_price, __pyx_v_high_price, __pyx_v_low_price, __pyx_v_volume, __pyx_v_closed, __pyx_v_timestamp, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_17candle(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_start, CYTHON_UNUSED PyObject *__pyx_v_stop, PyObject *__pyx_v_interval, CYTHON_UNUSED PyObject *__pyx_v_trades, PyObject *__pyx_v_open_price, PyObject *__pyx_v_close_price, PyObject *__pyx_v_high_price, PyObject *__pyx_v_low_price, PyObject *__pyx_v_volume, PyObject *__pyx_v_closed, PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("candle", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 172, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_start = __pyx_v_start;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_start);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_start);
  __pyx_cur_scope->__pyx_v_stop = __pyx_v_stop;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stop);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_stop);
  __pyx_cur_scope->__pyx_v_interval = __pyx_v_interval;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_interval);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_interval);
  __pyx_cur_scope->__pyx_v_trades = __pyx_v_trades;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_trades);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_trades);
  __pyx_cur_scope->__pyx_v_open_price = __pyx_v_open_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_open_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_open_price);
  __pyx_cur_scope->__pyx_v_close_price = __pyx_v_close_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_close_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_close_price);
  __pyx_cur_scope->__pyx_v_high_price = __pyx_v_high_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_high_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_high_price);
  __pyx_cur_scope->__pyx_v_low_price = __pyx_v_low_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_low_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_low_price);
  __pyx_cur_scope->__pyx_v_volume = __pyx_v_volume;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_volume);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_volume);
  __pyx_cur_scope->__pyx_v_closed = __pyx_v_closed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_closed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_closed);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_19generator3, __pyx_codeobj__5, (PyObject *) __pyx_cur_scope, __pyx_n_s_candle, __pyx_n_s_CryptofeedWebsocketConnector_can, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.candle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_19generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("candle", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_await;
    case 2: goto __pyx_L7_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 172, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":174
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             time_frame = commons_enums.TimeFrames(interval)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_symbol); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":175
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)             # <<<<<<<<<<<<<<
 *             time_frame = commons_enums.TimeFrames(interval)
 *             candle = [
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_symbol))||((__pyx_cur_scope->__pyx_v_symbol) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_cur_scope->__pyx_v_symbol)->tp_name), 0))) __PYX_ERR(0, 175, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->get_pair_from_exchange(__pyx_cur_scope->__pyx_v_self, ((PyObject*)__pyx_cur_scope->__pyx_v_symbol), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_symbol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_symbol, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":176
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 *             time_frame = commons_enums.TimeFrames(interval)             # <<<<<<<<<<<<<<
 *             candle = [
 *                 timestamp,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_commons_enums); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_TimeFrames); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_cur_scope->__pyx_v_interval) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_interval);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v_time_frame = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":179
 *             candle = [
 *                 timestamp,
 *                 float(open_price),             # <<<<<<<<<<<<<<
 *                 float(high_price),
 *                 float(low_price),
 */
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_open_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":180
 *                 timestamp,
 *                 float(open_price),
 *                 float(high_price),             # <<<<<<<<<<<<<<
 *                 float(low_price),
 *                 float(close_price),
 */
    __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_high_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":181
 *                 float(open_price),
 *                 float(high_price),
 *                 float(low_price),             # <<<<<<<<<<<<<<
 *                 float(close_price),
 *                 float(volume),
 */
    __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_low_price); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":182
 *                 float(high_price),
 *                 float(low_price),
 *                 float(close_price),             # <<<<<<<<<<<<<<
 *                 float(volume),
 *             ]
 */
    __pyx_t_5 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_close_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":183
 *                 float(low_price),
 *                 float(close_price),
 *                 float(volume),             # <<<<<<<<<<<<<<
 *             ]
 *             if not closed:
 */
    __pyx_t_6 = __Pyx_PyNumber_Float(__pyx_cur_scope->__pyx_v_volume); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":177
 *             symbol = self.get_pair_from_exchange(symbol)
 *             time_frame = commons_enums.TimeFrames(interval)
 *             candle = [             # <<<<<<<<<<<<<<
 *                 timestamp,
 *                 float(open_price),
 */
    __pyx_t_7 = PyList_New(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
    PyList_SET_ITEM(__pyx_t_7, 0, __pyx_cur_scope->__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_7, 3, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_7, 4, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyList_SET_ITEM(__pyx_t_7, 5, __pyx_t_6);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_cur_scope->__pyx_v_candle = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":185
 *                 float(volume),
 *             ]
 *             if not closed:             # <<<<<<<<<<<<<<
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,
 *                                            time_frame=time_frame,
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_closed); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
    __pyx_t_8 = ((!__pyx_t_1) != 0);
    if (__pyx_t_8) {

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":186
 *             ]
 *             if not closed:
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,             # <<<<<<<<<<<<<<
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_push_to_channel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_trading_constants); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_KLINE_CHANNEL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":187
 *             if not closed:
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,
 *                                            time_frame=time_frame,             # <<<<<<<<<<<<<<
 *                                            symbol=symbol,
 *                                            kline=candle)
 */
      __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_time_frame, __pyx_cur_scope->__pyx_v_time_frame) < 0) __PYX_ERR(0, 187, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":188
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,             # <<<<<<<<<<<<<<
 *                                            kline=candle)
 *             else:
 */
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_symbol, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 187, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":189
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 *                                            kline=candle)             # <<<<<<<<<<<<<<
 *             else:
 *                 await self.push_to_channel(trading_constants.OHLCV_CHANNEL,
 */
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_kline, __pyx_cur_scope->__pyx_v_candle) < 0) __PYX_ERR(0, 187, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":186
 *             ]
 *             if not closed:
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,             # <<<<<<<<<<<<<<
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L6_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 186, __pyx_L1_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 186, __pyx_L1_error)
        }
      }

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":185
 *                 float(volume),
 *             ]
 *             if not closed:             # <<<<<<<<<<<<<<
 *                 await self.push_to_channel(trading_constants.KLINE_CHANNEL,
 *                                            time_frame=time_frame,
 */
      goto __pyx_L5;
    }

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":191
 *                                            kline=candle)
 *             else:
 *                 await self.push_to_channel(trading_constants.OHLCV_CHANNEL,             # <<<<<<<<<<<<<<
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_push_to_channel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_trading_constants); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_OHLCV_CHANNEL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":192
 *             else:
 *                 await self.push_to_channel(trading_constants.OHLCV_CHANNEL,
 *                                            time_frame=time_frame,             # <<<<<<<<<<<<<<
 *                                            symbol=symbol,
 *                                            candle=candle)
 */
      __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_time_frame, __pyx_cur_scope->__pyx_v_time_frame) < 0) __PYX_ERR(0, 192, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":193
 *                 await self.push_to_channel(trading_constants.OHLCV_CHANNEL,
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,             # <<<<<<<<<<<<<<
 *                                            candle=candle)
 * 
 */
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_symbol, __pyx_cur_scope->__pyx_v_symbol) < 0) __PYX_ERR(0, 192, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":194
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 *                                            candle=candle)             # <<<<<<<<<<<<<<
 * 
 *     async def delta(self, feed, symbol, delta, timestamp, receipt_timestamp):
 */
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_candle, __pyx_cur_scope->__pyx_v_candle) < 0) __PYX_ERR(0, 192, __pyx_L1_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":191
 *                                            kline=candle)
 *             else:
 *                 await self.push_to_channel(trading_constants.OHLCV_CHANNEL,             # <<<<<<<<<<<<<<
 *                                            time_frame=time_frame,
 *                                            symbol=symbol,
 */
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L7_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 191, __pyx_L1_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 191, __pyx_L1_error)
        }
      }
    }
    __pyx_L5:;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":174
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:             # <<<<<<<<<<<<<<
 *             symbol = self.get_pair_from_exchange(symbol)
 *             time_frame = commons_enums.TimeFrames(interval)
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":172
 *         ]
 * 
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,             # <<<<<<<<<<<<<<
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("candle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_22generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":196
 *                                            candle=candle)
 * 
 *     async def delta(self, feed, symbol, delta, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta = {"delta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  CYTHON_UNUSED PyObject *__pyx_v_symbol = 0;
  CYTHON_UNUSED PyObject *__pyx_v_delta = 0;
  CYTHON_UNUSED PyObject *__pyx_v_timestamp = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("delta (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_delta,&__pyx_n_s_timestamp,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("delta", 1, 5, 5, 1); __PYX_ERR(0, 196, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("delta", 1, 5, 5, 2); __PYX_ERR(0, 196, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("delta", 1, 5, 5, 3); __PYX_ERR(0, 196, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("delta", 1, 5, 5, 4); __PYX_ERR(0, 196, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "delta") < 0)) __PYX_ERR(0, 196, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_delta = values[2];
    __pyx_v_timestamp = values[3];
    __pyx_v_receipt_timestamp = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("delta", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 196, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_20delta(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_delta, __pyx_v_timestamp, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_20delta(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_delta, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delta", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 196, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_delta = __pyx_v_delta;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_delta);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_delta);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_22generator4, __pyx_codeobj__6, (PyObject *) __pyx_cur_scope, __pyx_n_s_delta, __pyx_n_s_CryptofeedWebsocketConnector_del, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_22generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("delta", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 196, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_25generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":199
 *         pass
 * 
 *     async def liquidations(self, feed, symbol, side, leaves_qty, price, order_id, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations = {"liquidations", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  CYTHON_UNUSED PyObject *__pyx_v_symbol = 0;
  CYTHON_UNUSED PyObject *__pyx_v_side = 0;
  CYTHON_UNUSED PyObject *__pyx_v_leaves_qty = 0;
  CYTHON_UNUSED PyObject *__pyx_v_price = 0;
  CYTHON_UNUSED PyObject *__pyx_v_order_id = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("liquidations (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_side,&__pyx_n_s_leaves_qty,&__pyx_n_s_price,&__pyx_n_s_order_id,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 1); __PYX_ERR(0, 199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 2); __PYX_ERR(0, 199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_qty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 3); __PYX_ERR(0, 199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_price)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 4); __PYX_ERR(0, 199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 5); __PYX_ERR(0, 199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, 6); __PYX_ERR(0, 199, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "liquidations") < 0)) __PYX_ERR(0, 199, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_side = values[2];
    __pyx_v_leaves_qty = values[3];
    __pyx_v_price = values[4];
    __pyx_v_order_id = values[5];
    __pyx_v_receipt_timestamp = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("liquidations", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 199, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.liquidations", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_23liquidations(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_side, __pyx_v_leaves_qty, __pyx_v_price, __pyx_v_order_id, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_23liquidations(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_side, CYTHON_UNUSED PyObject *__pyx_v_leaves_qty, CYTHON_UNUSED PyObject *__pyx_v_price, CYTHON_UNUSED PyObject *__pyx_v_order_id, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("liquidations", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 199, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_side = __pyx_v_side;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_side);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_side);
  __pyx_cur_scope->__pyx_v_leaves_qty = __pyx_v_leaves_qty;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_leaves_qty);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_leaves_qty);
  __pyx_cur_scope->__pyx_v_price = __pyx_v_price;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_price);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_price);
  __pyx_cur_scope->__pyx_v_order_id = __pyx_v_order_id;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_order_id);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_order_id);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_25generator5, __pyx_codeobj__7, (PyObject *) __pyx_cur_scope, __pyx_n_s_liquidations, __pyx_n_s_CryptofeedWebsocketConnector_liq, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.liquidations", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_25generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("liquidations", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 199, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("liquidations", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_28generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":202
 *         pass
 * 
 *     async def funding(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding = {"funding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("funding (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("funding", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "funding", 1))) return NULL;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_26funding(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_26funding(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("funding", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 202, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_XINCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_28generator6, __pyx_codeobj__8, (PyObject *) __pyx_cur_scope, __pyx_n_s_funding, __pyx_n_s_CryptofeedWebsocketConnector_fun, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.funding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_28generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("funding", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("funding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_31generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":205
 *         pass
 * 
 *     async def open_interest(self, feed, symbol, open_interest, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest = {"open_interest", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_feed = 0;
  CYTHON_UNUSED PyObject *__pyx_v_symbol = 0;
  CYTHON_UNUSED PyObject *__pyx_v_open_interest = 0;
  CYTHON_UNUSED PyObject *__pyx_v_timestamp = 0;
  CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open_interest (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feed,&__pyx_n_s_symbol,&__pyx_n_s_open_interest,&__pyx_n_s_timestamp,&__pyx_n_s_receipt_timestamp,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feed)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open_interest", 1, 5, 5, 1); __PYX_ERR(0, 205, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_open_interest)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open_interest", 1, 5, 5, 2); __PYX_ERR(0, 205, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open_interest", 1, 5, 5, 3); __PYX_ERR(0, 205, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_receipt_timestamp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("open_interest", 1, 5, 5, 4); __PYX_ERR(0, 205, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "open_interest") < 0)) __PYX_ERR(0, 205, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_feed = values[0];
    __pyx_v_symbol = values[1];
    __pyx_v_open_interest = values[2];
    __pyx_v_timestamp = values[3];
    __pyx_v_receipt_timestamp = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open_interest", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 205, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.open_interest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_29open_interest(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_feed, __pyx_v_symbol, __pyx_v_open_interest, __pyx_v_timestamp, __pyx_v_receipt_timestamp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_29open_interest(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_feed, CYTHON_UNUSED PyObject *__pyx_v_symbol, CYTHON_UNUSED PyObject *__pyx_v_open_interest, CYTHON_UNUSED PyObject *__pyx_v_timestamp, CYTHON_UNUSED PyObject *__pyx_v_receipt_timestamp) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open_interest", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 205, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_feed = __pyx_v_feed;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_feed);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_feed);
  __pyx_cur_scope->__pyx_v_symbol = __pyx_v_symbol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_symbol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_symbol);
  __pyx_cur_scope->__pyx_v_open_interest = __pyx_v_open_interest;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_open_interest);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_open_interest);
  __pyx_cur_scope->__pyx_v_timestamp = __pyx_v_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timestamp);
  __pyx_cur_scope->__pyx_v_receipt_timestamp = __pyx_v_receipt_timestamp;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_receipt_timestamp);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_31generator7, __pyx_codeobj__9, (PyObject *) __pyx_cur_scope, __pyx_n_s_open_interest, __pyx_n_s_CryptofeedWebsocketConnector_ope, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 205, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.open_interest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_31generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open_interest", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 205, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("open_interest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_34generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":208
 *         pass
 * 
 *     async def volume(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume = {"volume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("volume (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("volume", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "volume", 1))) return NULL;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_32volume(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_32volume(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("volume", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 208, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_XINCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_34generator8, __pyx_codeobj__10, (PyObject *) __pyx_cur_scope, __pyx_n_s_volume, __pyx_n_s_CryptofeedWebsocketConnector_vol, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.volume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_34generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("volume", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 208, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("volume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_37generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":211
 *         pass
 * 
 *     async def futures_index(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index = {"futures_index", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("futures_index (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("futures_index", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "futures_index", 1))) return NULL;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_35futures_index(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_35futures_index(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("futures_index", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 211, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_XINCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_37generator9, __pyx_codeobj__11, (PyObject *) __pyx_cur_scope, __pyx_n_s_futures_index, __pyx_n_s_CryptofeedWebsocketConnector_fut, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.futures_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_37generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("futures_index", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 211, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("futures_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_40generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":214
 *         pass
 * 
 *     async def market_info(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Coingecko only
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info = {"market_info", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("market_info (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("market_info", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "market_info", 1))) return NULL;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_38market_info(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_38market_info(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("market_info", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 214, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_XINCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_40generator10, __pyx_codeobj__12, (PyObject *) __pyx_cur_scope, __pyx_n_s_market_info, __pyx_n_s_CryptofeedWebsocketConnector_mar, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.market_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_40generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("market_info", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 214, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("market_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_43generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":217
 *         pass  # Coingecko only
 * 
 *     async def transactions(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Whale alert only
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions = {"transactions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("transactions (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("transactions", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "transactions", 1))) return NULL;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_41transactions(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_41transactions(CYTHON_UNUSED struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("transactions", 0);
  __pyx_cur_scope = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 217, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_XINCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_43generator11, __pyx_codeobj__13, (PyObject *) __pyx_cur_scope, __pyx_n_s_transactions, __pyx_n_s_CryptofeedWebsocketConnector_tra_2, __pyx_n_s_octobot_trading_exchanges_connec); if (unlikely(!gen)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.transactions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_43generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *__pyx_cur_scope = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("transactions", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 217, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("transactions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":220
 *         pass  # Whale alert only
 * 
 *     def get_pair_from_exchange(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */

static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange(PyObject *__pyx_v_self, PyObject *__pyx_v_pair); /*proto*/
static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_pair_from_exchange(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pair_from_exchange", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pair_from_exchange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_pair) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_pair);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 220, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":221
 * 
 *     def get_pair_from_exchange(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":222
 *     def get_pair_from_exchange(self, pair) -> str:
 *         try:
 *             return symbol_util.convert_symbol(             # <<<<<<<<<<<<<<
 *                 symbol=pair,
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_symbol_util); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_convert_symbol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":223
 *         try:
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,             # <<<<<<<<<<<<<<
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 *                 new_symbol_separator=octobot_commons.MARKET_SEPARATOR,
 */
      __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_symbol, __pyx_v_pair) < 0) __PYX_ERR(0, 223, __pyx_L3_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":224
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,             # <<<<<<<<<<<<<<
 *                 new_symbol_separator=octobot_commons.MARKET_SEPARATOR,
 *                 should_uppercase=True,
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_CRYPTOFEED_DEFAULT_MARKET_SEPARA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 224, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_symbol_separator, __pyx_t_3) < 0) __PYX_ERR(0, 223, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":225
 *                 symbol=pair,
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 *                 new_symbol_separator=octobot_commons.MARKET_SEPARATOR,             # <<<<<<<<<<<<<<
 *                 should_uppercase=True,
 *             )
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_octobot_commons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 225, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET_SEPARATOR); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_new_symbol_separator, __pyx_t_4) < 0) __PYX_ERR(0, 223, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":226
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 *                 new_symbol_separator=octobot_commons.MARKET_SEPARATOR,
 *                 should_uppercase=True,             # <<<<<<<<<<<<<<
 *             )
 *         except Exception:
 */
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_should_uppercase, Py_True) < 0) __PYX_ERR(0, 223, __pyx_L3_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":222
 *     def get_pair_from_exchange(self, pair) -> str:
 *         try:
 *             return symbol_util.convert_symbol(             # <<<<<<<<<<<<<<
 *                 symbol=pair,
 *                 symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 222, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 222, __pyx_L3_error)
      __pyx_r = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":221
 * 
 *     def get_pair_from_exchange(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":228
 *                 should_uppercase=True,
 *             )
 *         except Exception:             # <<<<<<<<<<<<<<
 *             self.logger.error(f"Failed to get market of {pair}")
 *         return ""
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_pair_from_exchange", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 228, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":229
 *             )
 *         except Exception:
 *             self.logger.error(f"Failed to get market of {pair}")             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.logger, __pyx_n_s_error); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_v_pair); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 229, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Failed_to_get_market_of, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 229, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 229, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":221
 * 
 *     def get_pair_from_exchange(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":230
 *         except Exception:
 *             self.logger.error(f"Failed to get market of {pair}")
 *         return ""             # <<<<<<<<<<<<<<
 * 
 *     def get_exchange_pair(self, pair) -> str:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u__14);
  __pyx_r = __pyx_kp_u__14;
  goto __pyx_L0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":220
 *         pass  # Whale alert only
 * 
 *     def get_pair_from_exchange(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_pair_from_exchange", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange(PyObject *__pyx_v_self, PyObject *__pyx_v_pair); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange = {"get_pair_from_exchange", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange, METH_O, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange(PyObject *__pyx_v_self, PyObject *__pyx_v_pair) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pair_from_exchange (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pair), (&PyUnicode_Type), 1, "pair", 1))) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_44get_pair_from_exchange(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), ((PyObject*)__pyx_v_pair));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_44get_pair_from_exchange(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pair_from_exchange", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_pair_from_exchange(__pyx_v_self, __pyx_v_pair, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_pair_from_exchange", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":232
 *         return ""
 * 
 *     def get_exchange_pair(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */

static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair(PyObject *__pyx_v_self, PyObject *__pyx_v_pair); /*proto*/
static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_exchange_pair(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_exchange_pair", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_exchange_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_pair) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_pair);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 232, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":233
 * 
 *     def get_exchange_pair(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":234
 *     def get_exchange_pair(self, pair) -> str:
 *         try:
 *             return symbol_util.convert_symbol(             # <<<<<<<<<<<<<<
 *                 symbol=pair,
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_symbol_util); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_convert_symbol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":235
 *         try:
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,             # <<<<<<<<<<<<<<
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,
 *                 new_symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 */
      __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_symbol, __pyx_v_pair) < 0) __PYX_ERR(0, 235, __pyx_L3_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":236
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,             # <<<<<<<<<<<<<<
 *                 new_symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 *                 should_uppercase=True,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_octobot_commons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET_SEPARATOR); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 236, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_symbol_separator, __pyx_t_4) < 0) __PYX_ERR(0, 235, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":237
 *                 symbol=pair,
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,
 *                 new_symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,             # <<<<<<<<<<<<<<
 *                 should_uppercase=True,
 *             )
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_CRYPTOFEED_DEFAULT_MARKET_SEPARA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 237, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_new_symbol_separator, __pyx_t_4) < 0) __PYX_ERR(0, 235, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":238
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,
 *                 new_symbol_separator=self.CRYPTOFEED_DEFAULT_MARKET_SEPARATOR,
 *                 should_uppercase=True,             # <<<<<<<<<<<<<<
 *             )
 *         except Exception:
 */
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_should_uppercase, Py_True) < 0) __PYX_ERR(0, 235, __pyx_L3_error)

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":234
 *     def get_exchange_pair(self, pair) -> str:
 *         try:
 *             return symbol_util.convert_symbol(             # <<<<<<<<<<<<<<
 *                 symbol=pair,
 *                 symbol_separator=octobot_commons.MARKET_SEPARATOR,
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 234, __pyx_L3_error)
      __pyx_r = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":233
 * 
 *     def get_exchange_pair(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":240
 *                 should_uppercase=True,
 *             )
 *         except Exception:             # <<<<<<<<<<<<<<
 *             self.logger.error(f"Failed to get market of {pair}")
 *         return ""
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_exchange_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 240, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);

      /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":241
 *             )
 *         except Exception:
 *             self.logger.error(f"Failed to get market of {pair}")             # <<<<<<<<<<<<<<
 *         return ""
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.logger, __pyx_n_s_error); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 241, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_v_pair); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 241, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Failed_to_get_market_of, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 241, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 241, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":233
 * 
 *     def get_exchange_pair(self, pair) -> str:
 *         try:             # <<<<<<<<<<<<<<
 *             return symbol_util.convert_symbol(
 *                 symbol=pair,
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
  }

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":242
 *         except Exception:
 *             self.logger.error(f"Failed to get market of {pair}")
 *         return ""             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u__14);
  __pyx_r = __pyx_kp_u__14;
  goto __pyx_L0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":232
 *         return ""
 * 
 *     def get_exchange_pair(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_exchange_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair(PyObject *__pyx_v_self, PyObject *__pyx_v_pair); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair = {"get_exchange_pair", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair, METH_O, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair(PyObject *__pyx_v_self, PyObject *__pyx_v_pair) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_exchange_pair (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pair), (&PyUnicode_Type), 1, "pair", 1))) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_46get_exchange_pair(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), ((PyObject*)__pyx_v_pair));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_46get_exchange_pair(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_pair) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_exchange_pair", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_exchange_pair(__pyx_v_self, __pyx_v_pair, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.get_exchange_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.pxd":20
 * 
 * cdef class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
 *     cdef public dict callback_by_feed             # <<<<<<<<<<<<<<
 * 
 *     cpdef str get_pair_from_exchange(self, str pair)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed___get__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed___get__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->callback_by_feed);
  __pyx_r = __pyx_v_self->callback_by_feed;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_2__set__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_2__set__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->callback_by_feed);
  __Pyx_DECREF(__pyx_v_self->callback_by_feed);
  __pyx_v_self->callback_by_feed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.callback_by_feed.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_4__del__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_4__del__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->callback_by_feed);
  __Pyx_DECREF(__pyx_v_self->callback_by_feed);
  __pyx_v_self->callback_by_feed = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__ = {"__reduce_cython__", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__, METH_NOARGS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_48__reduce_cython__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_48__reduce_cython__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.books, self.bot_mainloop, self.callback_by_feed, self.channels, self.client, self.config, self.currencies, self.exchange, self.exchange_id, self.exchange_manager, self.logger, self.name, self.pairs, self.time_frames)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(14); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.books);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.books);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.books);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.bot_mainloop);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.bot_mainloop);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.bot_mainloop);
  __Pyx_INCREF(__pyx_v_self->callback_by_feed);
  __Pyx_GIVEREF(__pyx_v_self->callback_by_feed);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->callback_by_feed);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.channels);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.channels);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.channels);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.client);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.client);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->__pyx_base.client);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.config);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.config);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->__pyx_base.config);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.currencies);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.currencies);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->__pyx_base.currencies);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.exchange);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.exchange);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_self->__pyx_base.exchange);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.exchange_id);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.exchange_id);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_v_self->__pyx_base.exchange_id);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.exchange_manager));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.exchange_manager));
  PyTuple_SET_ITEM(__pyx_t_1, 9, ((PyObject *)__pyx_v_self->__pyx_base.exchange_manager));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.logger);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.logger);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_v_self->__pyx_base.logger);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_v_self->__pyx_base.name);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.pairs);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pairs);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_v_self->__pyx_base.pairs);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.time_frames);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.time_frames);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_v_self->__pyx_base.time_frames);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.books, self.bot_mainloop, self.callback_by_feed, self.channels, self.client, self.config, self.currencies, self.exchange, self.exchange_id, self.exchange_manager, self.logger, self.name, self.pairs, self.time_frames)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.books, self.bot_mainloop, self.callback_by_feed, self.channels, self.client, self.config, self.currencies, self.exchange, self.exchange_id, self.exchange_manager, self.logger, self.name, self.pairs, self.time_frames)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.books is not None or self.bot_mainloop is not None or self.callback_by_feed is not None or self.channels is not None or self.client is not None or self.config is not None or self.currencies is not None or self.exchange is not None or self.exchange_id is not None or self.exchange_manager is not None or self.logger is not None or self.name is not None or self.pairs is not None or self.time_frames is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.books, self.bot_mainloop, self.callback_by_feed, self.channels, self.client, self.config, self.currencies, self.exchange, self.exchange_id, self.exchange_manager, self.logger, self.name, self.pairs, self.time_frames)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.books is not None or self.bot_mainloop is not None or self.callback_by_feed is not None or self.channels is not None or self.client is not None or self.config is not None or self.currencies is not None or self.exchange is not None or self.exchange_id is not None or self.exchange_manager is not None or self.logger is not None or self.name is not None or self.pairs is not None or self.time_frames is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->__pyx_base.books != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.bot_mainloop != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->callback_by_feed != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.channels != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->__pyx_base.client != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.config != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->__pyx_base.currencies != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.exchange != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->__pyx_base.exchange_id != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.exchange_manager) != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->__pyx_base.logger != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.name != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->__pyx_base.pairs != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.time_frames != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.books is not None or self.bot_mainloop is not None or self.callback_by_feed is not None or self.channels is not None or self.client is not None or self.config is not None or self.currencies is not None or self.exchange is not None or self.exchange_id is not None or self.exchange_manager is not None or self.logger is not None or self.name is not None or self.pairs is not None or self.time_frames is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.books is not None or self.bot_mainloop is not None or self.callback_by_feed is not None or self.channels is not None or self.client is not None or self.config is not None or self.currencies is not None or self.exchange is not None or self.exchange_id is not None or self.exchange_manager is not None or self.logger is not None or self.name is not None or self.pairs is not None or self.time_frames is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CryptofeedWebsock); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_236055009);
    __Pyx_GIVEREF(__pyx_int_236055009);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_236055009);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.books is not None or self.bot_mainloop is not None or self.callback_by_feed is not None or self.channels is not None or self.client is not None or self.config is not None or self.currencies is not None or self.exchange is not None or self.exchange_id is not None or self.exchange_manager is not None or self.logger is not None or self.name is not None or self.pairs is not None or self.time_frames is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, None), state
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_CryptofeedWebsock); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_236055009);
    __Pyx_GIVEREF(__pyx_int_236055009);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_236055009);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__ = {"__setstate_cython__", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__, METH_O, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_50__setstate_cython__(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_50__setstate_cython__(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CryptofeedWebsocketConnector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_1__pyx_unpickle_CryptofeedWebsocketConnector(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_1__pyx_unpickle_CryptofeedWebsocketConnector = {"__pyx_unpickle_CryptofeedWebsocketConnector", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_1__pyx_unpickle_CryptofeedWebsocketConnector, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_1__pyx_unpickle_CryptofeedWebsocketConnector(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CryptofeedWebsocketConnector (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CryptofeedWebsocketConnector", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CryptofeedWebsocketConnector", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_CryptofeedWebsocketConnector") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CryptofeedWebsocketConnector", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_unpickle_CryptofeedWebsocketConnector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CryptofeedWebsocketConnector", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xe11e9e1:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xe11e9e1) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xe11e9e1:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xe11e9e1:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xe1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xe11e9e1:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector__set_state(((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xe11e9e1 = (books, bot_mainloop, callback_by_feed, channels, client, config, currencies, exchange, exchange_id, exchange_manager, logger, name, pairs, time_frames))" % __pyx_checksum)
 *     __pyx_result = CryptofeedWebsocketConnector.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CryptofeedWebsocketConnector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_unpickle_CryptofeedWebsocketConnector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_unpickle_CryptofeedWebsocketConnector__set_state(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CryptofeedWebsocketConnector__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.books);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.books);
  __pyx_v___pyx_result->__pyx_base.books = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.bot_mainloop);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.bot_mainloop);
  __pyx_v___pyx_result->__pyx_base.bot_mainloop = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->callback_by_feed);
  __Pyx_DECREF(__pyx_v___pyx_result->callback_by_feed);
  __pyx_v___pyx_result->callback_by_feed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.channels);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.channels);
  __pyx_v___pyx_result->__pyx_base.channels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.client);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.client);
  __pyx_v___pyx_result->__pyx_base.client = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.config);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.config);
  __pyx_v___pyx_result->__pyx_base.config = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.currencies);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.currencies);
  __pyx_v___pyx_result->__pyx_base.currencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.exchange);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.exchange);
  __pyx_v___pyx_result->__pyx_base.exchange = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.exchange_id);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.exchange_id);
  __pyx_v___pyx_result->__pyx_base.exchange_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.exchange_manager);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.exchange_manager));
  __pyx_v___pyx_result->__pyx_base.exchange_manager = ((struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.logger);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.logger);
  __pyx_v___pyx_result->__pyx_base.logger = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.pairs);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.pairs);
  __pyx_v___pyx_result->__pyx_base.pairs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.time_frames);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.time_frames);
  __pyx_v___pyx_result->__pyx_base.time_frames = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 14) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CryptofeedWebsocketConnector__set_state(<CryptofeedWebsocketConnector> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CryptofeedWebsocketConnector__set_state(CryptofeedWebsocketConnector __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.books = __pyx_state[0]; __pyx_result.bot_mainloop = __pyx_state[1]; __pyx_result.callback_by_feed = __pyx_state[2]; __pyx_result.channels = __pyx_state[3]; __pyx_result.client = __pyx_state[4]; __pyx_result.config = __pyx_state[5]; __pyx_result.currencies = __pyx_state[6]; __pyx_result.exchange = __pyx_state[7]; __pyx_result.exchange_id = __pyx_state[8]; __pyx_result.exchange_manager = __pyx_state[9]; __pyx_result.logger = __pyx_state[10]; __pyx_result.name = __pyx_state[11]; __pyx_result.pairs = __pyx_state[12]; __pyx_result.time_frames = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_unpickle_CryptofeedWebsocketConnector__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *p;
  PyObject *o = __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange*)__pyx_vtabptr_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;
  p->callback_by_feed = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->callback_by_feed);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange)) __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)o;
  e = ((likely(__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange)) ? ((__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_traverse) ? __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector)); if (e) return e;
  if (p->callback_by_feed) {
    e = (*v)(p->callback_by_feed, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *)o;
  if (likely(__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange)) { if (__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_clear) __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);
  tmp = ((PyObject*)p->callback_by_feed);
  p->callback_by_feed = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_callback_by_feed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_1__get__(o);
}

static int __pyx_setprop_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_callback_by_feed(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_3__set__(o, v);
  }
  else {
    return __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16callback_by_feed_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector[] = {
  {"get_feed_name", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name, METH_NOARGS, 0},
  {"fix_signal_handler", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler, METH_NOARGS, __pyx_doc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_4fix_signal_handler},
  {"ticker", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker, METH_VARARGS|METH_KEYWORDS, 0},
  {"trade", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade, METH_VARARGS|METH_KEYWORDS, 0},
  {"book", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book, METH_VARARGS|METH_KEYWORDS, 0},
  {"_convert_book_prices_to_orders", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders, METH_VARARGS|METH_KEYWORDS, 0},
  {"candle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle, METH_VARARGS|METH_KEYWORDS, 0},
  {"delta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta, METH_VARARGS|METH_KEYWORDS, 0},
  {"liquidations", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations, METH_VARARGS|METH_KEYWORDS, 0},
  {"funding", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding, METH_VARARGS|METH_KEYWORDS, 0},
  {"open_interest", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest, METH_VARARGS|METH_KEYWORDS, 0},
  {"volume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume, METH_VARARGS|METH_KEYWORDS, 0},
  {"futures_index", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index, METH_VARARGS|METH_KEYWORDS, 0},
  {"market_info", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info, METH_VARARGS|METH_KEYWORDS, 0},
  {"transactions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector[] = {
  {(char *)"callback_by_feed", __pyx_getprop_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_callback_by_feed, __pyx_setprop_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_callback_by_feed, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.CryptofeedWebsocketConnector", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_traverse*/
  __pyx_tp_clear_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ask);
  Py_CLEAR(p->__pyx_v_bid);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_timestamp);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker *)o;
  if (p->__pyx_v_ask) {
    e = (*v)(p->__pyx_v_ask, a); if (e) return e;
  }
  if (p->__pyx_v_bid) {
    e = (*v)(p->__pyx_v_bid, a); if (e) return e;
  }
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct__ticker", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_amount);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_order_id);
  Py_CLEAR(p->__pyx_v_price);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_side);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_timestamp);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade *)o;
  if (p->__pyx_v_amount) {
    e = (*v)(p->__pyx_v_amount, a); if (e) return e;
  }
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_order_id) {
    e = (*v)(p->__pyx_v_order_id, a); if (e) return e;
  }
  if (p->__pyx_v_price) {
    e = (*v)(p->__pyx_v_price, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_side) {
    e = (*v)(p->__pyx_v_side, a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_1_trade", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_book);
  Py_CLEAR(p->__pyx_v_book_instance);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_timestamp);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book *)o;
  if (p->__pyx_v_book) {
    e = (*v)(p->__pyx_v_book, a); if (e) return e;
  }
  if (p->__pyx_v_book_instance) {
    e = (*v)(p->__pyx_v_book_instance, a); if (e) return e;
  }
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_2_book", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_candle);
  Py_CLEAR(p->__pyx_v_close_price);
  Py_CLEAR(p->__pyx_v_closed);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_high_price);
  Py_CLEAR(p->__pyx_v_interval);
  Py_CLEAR(p->__pyx_v_low_price);
  Py_CLEAR(p->__pyx_v_open_price);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_start);
  Py_CLEAR(p->__pyx_v_stop);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_time_frame);
  Py_CLEAR(p->__pyx_v_timestamp);
  Py_CLEAR(p->__pyx_v_trades);
  Py_CLEAR(p->__pyx_v_volume);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle *)o;
  if (p->__pyx_v_candle) {
    e = (*v)(p->__pyx_v_candle, a); if (e) return e;
  }
  if (p->__pyx_v_close_price) {
    e = (*v)(p->__pyx_v_close_price, a); if (e) return e;
  }
  if (p->__pyx_v_closed) {
    e = (*v)(p->__pyx_v_closed, a); if (e) return e;
  }
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_high_price) {
    e = (*v)(p->__pyx_v_high_price, a); if (e) return e;
  }
  if (p->__pyx_v_interval) {
    e = (*v)(p->__pyx_v_interval, a); if (e) return e;
  }
  if (p->__pyx_v_low_price) {
    e = (*v)(p->__pyx_v_low_price, a); if (e) return e;
  }
  if (p->__pyx_v_open_price) {
    e = (*v)(p->__pyx_v_open_price, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_start) {
    e = (*v)(p->__pyx_v_start, a); if (e) return e;
  }
  if (p->__pyx_v_stop) {
    e = (*v)(p->__pyx_v_stop, a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_time_frame) {
    e = (*v)(p->__pyx_v_time_frame, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_trades) {
    e = (*v)(p->__pyx_v_trades, a); if (e) return e;
  }
  if (p->__pyx_v_volume) {
    e = (*v)(p->__pyx_v_volume, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_3_candle", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_delta);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_timestamp);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta *)o;
  if (p->__pyx_v_delta) {
    e = (*v)(p->__pyx_v_delta, a); if (e) return e;
  }
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_4_delta", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_leaves_qty);
  Py_CLEAR(p->__pyx_v_order_id);
  Py_CLEAR(p->__pyx_v_price);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_side);
  Py_CLEAR(p->__pyx_v_symbol);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations *)o;
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_leaves_qty) {
    e = (*v)(p->__pyx_v_leaves_qty, a); if (e) return e;
  }
  if (p->__pyx_v_order_id) {
    e = (*v)(p->__pyx_v_order_id, a); if (e) return e;
  }
  if (p->__pyx_v_price) {
    e = (*v)(p->__pyx_v_price, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_side) {
    e = (*v)(p->__pyx_v_side, a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_5_liquidations", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding *)o;
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_6_funding", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_feed);
  Py_CLEAR(p->__pyx_v_open_interest);
  Py_CLEAR(p->__pyx_v_receipt_timestamp);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_symbol);
  Py_CLEAR(p->__pyx_v_timestamp);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest *)o;
  if (p->__pyx_v_feed) {
    e = (*v)(p->__pyx_v_feed, a); if (e) return e;
  }
  if (p->__pyx_v_open_interest) {
    e = (*v)(p->__pyx_v_open_interest, a); if (e) return e;
  }
  if (p->__pyx_v_receipt_timestamp) {
    e = (*v)(p->__pyx_v_receipt_timestamp, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_symbol) {
    e = (*v)(p->__pyx_v_symbol, a); if (e) return e;
  }
  if (p->__pyx_v_timestamp) {
    e = (*v)(p->__pyx_v_timestamp, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_7_open_interest", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume *)o;
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_8_volume", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index *)o;
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_9_futures_index", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info *)o;
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_10_market_info", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions[8];
static int __pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions = 0;

static PyObject *__pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions)))) {
    o = (PyObject*)__pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions[--__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions];
    memset(o, 0, sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions(PyObject *o) {
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions)))) {
    __pyx_freelist_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions[__pyx_freecount_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions++] = ((struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *p = (struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions *)o;
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions = {
  PyVarObject_HEAD_INIT(0, 0)
  "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector.__pyx_scope_struct_11_transactions", /*tp_name*/
  sizeof(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cryptofeed_websocket_connector(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cryptofeed_websocket_connector},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cryptofeed_websocket_connector",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AMOUNT, __pyx_k_AMOUNT, sizeof(__pyx_k_AMOUNT), 0, 0, 1, 1},
  {&__pyx_n_s_ASK, __pyx_k_ASK, sizeof(__pyx_k_ASK), 0, 0, 1, 1},
  {&__pyx_n_s_BID, __pyx_k_BID, sizeof(__pyx_k_BID), 0, 0, 1, 1},
  {&__pyx_n_s_BOOK_DELTA, __pyx_k_BOOK_DELTA, sizeof(__pyx_k_BOOK_DELTA), 0, 0, 1, 1},
  {&__pyx_n_s_BUY, __pyx_k_BUY, sizeof(__pyx_k_BUY), 0, 0, 1, 1},
  {&__pyx_n_s_BookCallback, __pyx_k_BookCallback, sizeof(__pyx_k_BookCallback), 0, 0, 1, 1},
  {&__pyx_n_s_BookUpdateCallback, __pyx_k_BookUpdateCallback, sizeof(__pyx_k_BookUpdateCallback), 0, 0, 1, 1},
  {&__pyx_n_s_CANDLE, __pyx_k_CANDLE, sizeof(__pyx_k_CANDLE), 0, 0, 1, 1},
  {&__pyx_n_s_CANDLES, __pyx_k_CANDLES, sizeof(__pyx_k_CANDLES), 0, 0, 1, 1},
  {&__pyx_n_s_CRYPTOFEED_DEFAULT_MARKET_SEPARA, __pyx_k_CRYPTOFEED_DEFAULT_MARKET_SEPARA, sizeof(__pyx_k_CRYPTOFEED_DEFAULT_MARKET_SEPARA), 0, 0, 1, 1},
  {&__pyx_n_s_CandleCallback, __pyx_k_CandleCallback, sizeof(__pyx_k_CandleCallback), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector, __pyx_k_CryptofeedWebsocketConnector, sizeof(__pyx_k_CryptofeedWebsocketConnector), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector___r, __pyx_k_CryptofeedWebsocketConnector___r, sizeof(__pyx_k_CryptofeedWebsocketConnector___r), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector___s, __pyx_k_CryptofeedWebsocketConnector___s, sizeof(__pyx_k_CryptofeedWebsocketConnector___s), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector__co, __pyx_k_CryptofeedWebsocketConnector__co, sizeof(__pyx_k_CryptofeedWebsocketConnector__co), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_boo, __pyx_k_CryptofeedWebsocketConnector_boo, sizeof(__pyx_k_CryptofeedWebsocketConnector_boo), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_can, __pyx_k_CryptofeedWebsocketConnector_can, sizeof(__pyx_k_CryptofeedWebsocketConnector_can), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_del, __pyx_k_CryptofeedWebsocketConnector_del, sizeof(__pyx_k_CryptofeedWebsocketConnector_del), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_fix, __pyx_k_CryptofeedWebsocketConnector_fix, sizeof(__pyx_k_CryptofeedWebsocketConnector_fix), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_fun, __pyx_k_CryptofeedWebsocketConnector_fun, sizeof(__pyx_k_CryptofeedWebsocketConnector_fun), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_fut, __pyx_k_CryptofeedWebsocketConnector_fut, sizeof(__pyx_k_CryptofeedWebsocketConnector_fut), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_get, __pyx_k_CryptofeedWebsocketConnector_get, sizeof(__pyx_k_CryptofeedWebsocketConnector_get), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_get_2, __pyx_k_CryptofeedWebsocketConnector_get_2, sizeof(__pyx_k_CryptofeedWebsocketConnector_get_2), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_get_3, __pyx_k_CryptofeedWebsocketConnector_get_3, sizeof(__pyx_k_CryptofeedWebsocketConnector_get_3), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_liq, __pyx_k_CryptofeedWebsocketConnector_liq, sizeof(__pyx_k_CryptofeedWebsocketConnector_liq), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_mar, __pyx_k_CryptofeedWebsocketConnector_mar, sizeof(__pyx_k_CryptofeedWebsocketConnector_mar), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_ope, __pyx_k_CryptofeedWebsocketConnector_ope, sizeof(__pyx_k_CryptofeedWebsocketConnector_ope), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_tic, __pyx_k_CryptofeedWebsocketConnector_tic, sizeof(__pyx_k_CryptofeedWebsocketConnector_tic), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_tra, __pyx_k_CryptofeedWebsocketConnector_tra, sizeof(__pyx_k_CryptofeedWebsocketConnector_tra), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_tra_2, __pyx_k_CryptofeedWebsocketConnector_tra_2, sizeof(__pyx_k_CryptofeedWebsocketConnector_tra_2), 0, 0, 1, 1},
  {&__pyx_n_s_CryptofeedWebsocketConnector_vol, __pyx_k_CryptofeedWebsocketConnector_vol, sizeof(__pyx_k_CryptofeedWebsocketConnector_vol), 0, 0, 1, 1},
  {&__pyx_n_s_DEBUG, __pyx_k_DEBUG, sizeof(__pyx_k_DEBUG), 0, 0, 1, 1},
  {&__pyx_n_s_ECOBIC, __pyx_k_ECOBIC, sizeof(__pyx_k_ECOBIC), 0, 0, 1, 1},
  {&__pyx_n_s_EXCHANGE_FEEDS, __pyx_k_EXCHANGE_FEEDS, sizeof(__pyx_k_EXCHANGE_FEEDS), 0, 0, 1, 1},
  {&__pyx_n_s_ExchangeConstantsOrderBookInfoCo, __pyx_k_ExchangeConstantsOrderBookInfoCo, sizeof(__pyx_k_ExchangeConstantsOrderBookInfoCo), 0, 0, 1, 1},
  {&__pyx_n_s_ExchangeConstantsOrderColumns, __pyx_k_ExchangeConstantsOrderColumns, sizeof(__pyx_k_ExchangeConstantsOrderColumns), 0, 0, 1, 1},
  {&__pyx_n_s_FUNDING, __pyx_k_FUNDING, sizeof(__pyx_k_FUNDING), 0, 0, 1, 1},
  {&__pyx_n_s_FUTURES_INDEX, __pyx_k_FUTURES_INDEX, sizeof(__pyx_k_FUTURES_INDEX), 0, 0, 1, 1},
  {&__pyx_kp_u_Failed_to_get_market_of, __pyx_k_Failed_to_get_market_of, sizeof(__pyx_k_Failed_to_get_market_of), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_start_websocket_feed, __pyx_k_Failed_to_start_websocket_feed, sizeof(__pyx_k_Failed_to_start_websocket_feed), 0, 1, 0, 0},
  {&__pyx_n_s_FeedHandler, __pyx_k_FeedHandler, sizeof(__pyx_k_FeedHandler), 0, 0, 1, 1},
  {&__pyx_n_s_Feeds, __pyx_k_Feeds, sizeof(__pyx_k_Feeds), 0, 0, 1, 1},
  {&__pyx_n_s_FundingCallback, __pyx_k_FundingCallback, sizeof(__pyx_k_FundingCallback), 0, 0, 1, 1},
  {&__pyx_n_s_FuturesIndexCallback, __pyx_k_FuturesIndexCallback, sizeof(__pyx_k_FuturesIndexCallback), 0, 0, 1, 1},
  {&__pyx_n_s_ID, __pyx_k_ID, sizeof(__pyx_k_ID), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xe1, __pyx_k_Incompatible_checksums_s_vs_0xe1, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xe1), 0, 0, 1, 0},
  {&__pyx_n_s_KLINE_CHANNEL, __pyx_k_KLINE_CHANNEL, sizeof(__pyx_k_KLINE_CHANNEL), 0, 0, 1, 1},
  {&__pyx_n_s_L2_BOOK, __pyx_k_L2_BOOK, sizeof(__pyx_k_L2_BOOK), 0, 0, 1, 1},
  {&__pyx_n_s_L3_BOOK, __pyx_k_L3_BOOK, sizeof(__pyx_k_L3_BOOK), 0, 0, 1, 1},
  {&__pyx_n_s_LIQUIDATIONS, __pyx_k_LIQUIDATIONS, sizeof(__pyx_k_LIQUIDATIONS), 0, 0, 1, 1},
  {&__pyx_n_s_LOGGERS, __pyx_k_LOGGERS, sizeof(__pyx_k_LOGGERS), 0, 0, 1, 1},
  {&__pyx_n_s_LiquidationCallback, __pyx_k_LiquidationCallback, sizeof(__pyx_k_LiquidationCallback), 0, 0, 1, 1},
  {&__pyx_n_s_MARKET_INFO, __pyx_k_MARKET_INFO, sizeof(__pyx_k_MARKET_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_MARKET_SEPARATOR, __pyx_k_MARKET_SEPARATOR, sizeof(__pyx_k_MARKET_SEPARATOR), 0, 0, 1, 1},
  {&__pyx_n_s_MarketInfoCallback, __pyx_k_MarketInfoCallback, sizeof(__pyx_k_MarketInfoCallback), 0, 0, 1, 1},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_OHLCV_CHANNEL, __pyx_k_OHLCV_CHANNEL, sizeof(__pyx_k_OHLCV_CHANNEL), 0, 0, 1, 1},
  {&__pyx_n_s_OPEN_INTEREST, __pyx_k_OPEN_INTEREST, sizeof(__pyx_k_OPEN_INTEREST), 0, 0, 1, 1},
  {&__pyx_n_s_ORDER_BOOK_CHANNEL, __pyx_k_ORDER_BOOK_CHANNEL, sizeof(__pyx_k_ORDER_BOOK_CHANNEL), 0, 0, 1, 1},
  {&__pyx_n_s_ORDER_BOOK_TICKER_CHANNEL, __pyx_k_ORDER_BOOK_TICKER_CHANNEL, sizeof(__pyx_k_ORDER_BOOK_TICKER_CHANNEL), 0, 0, 1, 1},
  {&__pyx_n_s_OpenInterestCallback, __pyx_k_OpenInterestCallback, sizeof(__pyx_k_OpenInterestCallback), 0, 0, 1, 1},
  {&__pyx_n_s_PRICE, __pyx_k_PRICE, sizeof(__pyx_k_PRICE), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RECENT_TRADES_CHANNEL, __pyx_k_RECENT_TRADES_CHANNEL, sizeof(__pyx_k_RECENT_TRADES_CHANNEL), 0, 0, 1, 1},
  {&__pyx_n_s_SELL, __pyx_k_SELL, sizeof(__pyx_k_SELL), 0, 0, 1, 1},
  {&__pyx_n_s_SIDE, __pyx_k_SIDE, sizeof(__pyx_k_SIDE), 0, 0, 1, 1},
  {&__pyx_n_s_SIGNALS, __pyx_k_SIGNALS, sizeof(__pyx_k_SIGNALS), 0, 0, 1, 1},
  {&__pyx_n_s_SIZE, __pyx_k_SIZE, sizeof(__pyx_k_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_SYMBOL, __pyx_k_SYMBOL, sizeof(__pyx_k_SYMBOL), 0, 0, 1, 1},
  {&__pyx_n_s_TICKER, __pyx_k_TICKER, sizeof(__pyx_k_TICKER), 0, 0, 1, 1},
  {&__pyx_n_s_TIMESTAMP, __pyx_k_TIMESTAMP, sizeof(__pyx_k_TIMESTAMP), 0, 0, 1, 1},
  {&__pyx_n_s_TRADES, __pyx_k_TRADES, sizeof(__pyx_k_TRADES), 0, 0, 1, 1},
  {&__pyx_n_s_TRANSACTIONS, __pyx_k_TRANSACTIONS, sizeof(__pyx_k_TRANSACTIONS), 0, 0, 1, 1},
  {&__pyx_n_s_TYPE, __pyx_k_TYPE, sizeof(__pyx_k_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_TickerCallback, __pyx_k_TickerCallback, sizeof(__pyx_k_TickerCallback), 0, 0, 1, 1},
  {&__pyx_n_s_TimeFrames, __pyx_k_TimeFrames, sizeof(__pyx_k_TimeFrames), 0, 0, 1, 1},
  {&__pyx_n_s_TradeCallback, __pyx_k_TradeCallback, sizeof(__pyx_k_TradeCallback), 0, 0, 1, 1},
  {&__pyx_n_s_TradeOrderSide, __pyx_k_TradeOrderSide, sizeof(__pyx_k_TradeOrderSide), 0, 0, 1, 1},
  {&__pyx_n_s_TransactionsCallback, __pyx_k_TransactionsCallback, sizeof(__pyx_k_TransactionsCallback), 0, 0, 1, 1},
  {&__pyx_n_s_UNSUPPORTED, __pyx_k_UNSUPPORTED, sizeof(__pyx_k_UNSUPPORTED), 0, 0, 1, 1},
  {&__pyx_n_s_VOLUME, __pyx_k_VOLUME, sizeof(__pyx_k_VOLUME), 0, 0, 1, 1},
  {&__pyx_n_s_VolumeCallback, __pyx_k_VolumeCallback, sizeof(__pyx_k_VolumeCallback), 0, 0, 1, 1},
  {&__pyx_n_s_WebsocketFeeds, __pyx_k_WebsocketFeeds, sizeof(__pyx_k_WebsocketFeeds), 0, 0, 1, 1},
  {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
  {&__pyx_n_s__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 1, 1},
  {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
  {&__pyx_n_s_abstract_websocket, __pyx_k_abstract_websocket, sizeof(__pyx_k_abstract_websocket), 0, 0, 1, 1},
  {&__pyx_n_s_add_feed, __pyx_k_add_feed, sizeof(__pyx_k_add_feed), 0, 0, 1, 1},
  {&__pyx_n_s_amount, __pyx_k_amount, sizeof(__pyx_k_amount), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_ask, __pyx_k_ask, sizeof(__pyx_k_ask), 0, 0, 1, 1},
  {&__pyx_n_s_ask_price, __pyx_k_ask_price, sizeof(__pyx_k_ask_price), 0, 0, 1, 1},
  {&__pyx_n_s_ask_quantity, __pyx_k_ask_quantity, sizeof(__pyx_k_ask_quantity), 0, 0, 1, 1},
  {&__pyx_n_s_asks, __pyx_k_asks, sizeof(__pyx_k_asks), 0, 0, 1, 1},
  {&__pyx_n_s_asyncio, __pyx_k_asyncio, sizeof(__pyx_k_asyncio), 0, 0, 1, 1},
  {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
  {&__pyx_n_s_asyncio_tasks, __pyx_k_asyncio_tasks, sizeof(__pyx_k_asyncio_tasks), 0, 0, 1, 1},
  {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
  {&__pyx_n_s_bid, __pyx_k_bid, sizeof(__pyx_k_bid), 0, 0, 1, 1},
  {&__pyx_n_s_bid_price, __pyx_k_bid_price, sizeof(__pyx_k_bid_price), 0, 0, 1, 1},
  {&__pyx_n_s_bid_quantity, __pyx_k_bid_quantity, sizeof(__pyx_k_bid_quantity), 0, 0, 1, 1},
  {&__pyx_n_s_bids, __pyx_k_bids, sizeof(__pyx_k_bids), 0, 0, 1, 1},
  {&__pyx_n_s_book, __pyx_k_book, sizeof(__pyx_k_book), 0, 0, 1, 1},
  {&__pyx_n_s_book_instance, __pyx_k_book_instance, sizeof(__pyx_k_book_instance), 0, 0, 1, 1},
  {&__pyx_n_s_book_prices, __pyx_k_book_prices, sizeof(__pyx_k_book_prices), 0, 0, 1, 1},
  {&__pyx_n_s_book_side, __pyx_k_book_side, sizeof(__pyx_k_book_side), 0, 0, 1, 1},
  {&__pyx_n_s_callbacks, __pyx_k_callbacks, sizeof(__pyx_k_callbacks), 0, 0, 1, 1},
  {&__pyx_n_s_candle, __pyx_k_candle, sizeof(__pyx_k_candle), 0, 0, 1, 1},
  {&__pyx_n_s_candle_closed_only, __pyx_k_candle_closed_only, sizeof(__pyx_k_candle_closed_only), 0, 0, 1, 1},
  {&__pyx_n_s_candle_interval, __pyx_k_candle_interval, sizeof(__pyx_k_candle_interval), 0, 0, 1, 1},
  {&__pyx_n_s_channels, __pyx_k_channels, sizeof(__pyx_k_channels), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_close_price, __pyx_k_close_price, sizeof(__pyx_k_close_price), 0, 0, 1, 1},
  {&__pyx_n_s_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_commons_enums, __pyx_k_commons_enums, sizeof(__pyx_k_commons_enums), 0, 0, 1, 1},
  {&__pyx_n_s_commons_logging, __pyx_k_commons_logging, sizeof(__pyx_k_commons_logging), 0, 0, 1, 1},
  {&__pyx_n_s_config, __pyx_k_config, sizeof(__pyx_k_config), 0, 0, 1, 1},
  {&__pyx_n_s_convert_book_prices_to_orders, __pyx_k_convert_book_prices_to_orders, sizeof(__pyx_k_convert_book_prices_to_orders), 0, 0, 1, 1},
  {&__pyx_n_s_convert_symbol, __pyx_k_convert_symbol, sizeof(__pyx_k_convert_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_cryptofeed, __pyx_k_cryptofeed, sizeof(__pyx_k_cryptofeed), 0, 0, 1, 1},
  {&__pyx_n_s_cryptofeed_callback, __pyx_k_cryptofeed_callback, sizeof(__pyx_k_cryptofeed_callback), 0, 0, 1, 1},
  {&__pyx_n_s_cryptofeed_callbacks, __pyx_k_cryptofeed_callbacks, sizeof(__pyx_k_cryptofeed_callbacks), 0, 0, 1, 1},
  {&__pyx_n_s_cryptofeed_constants, __pyx_k_cryptofeed_constants, sizeof(__pyx_k_cryptofeed_constants), 0, 0, 1, 1},
  {&__pyx_n_s_cryptofeed_defines, __pyx_k_cryptofeed_defines, sizeof(__pyx_k_cryptofeed_defines), 0, 0, 1, 1},
  {&__pyx_n_s_delta, __pyx_k_delta, sizeof(__pyx_k_delta), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_exchange_manager, __pyx_k_exchange_manager, sizeof(__pyx_k_exchange_manager), 0, 0, 1, 1},
  {&__pyx_n_s_feed, __pyx_k_feed, sizeof(__pyx_k_feed), 0, 0, 1, 1},
  {&__pyx_n_s_feedhandler, __pyx_k_feedhandler, sizeof(__pyx_k_feedhandler), 0, 0, 1, 1},
  {&__pyx_n_u_feedhandler, __pyx_k_feedhandler, sizeof(__pyx_k_feedhandler), 0, 1, 0, 1},
  {&__pyx_n_s_fix_signal_handler, __pyx_k_fix_signal_handler, sizeof(__pyx_k_fix_signal_handler), 0, 0, 1, 1},
  {&__pyx_n_s_funding, __pyx_k_funding, sizeof(__pyx_k_funding), 0, 0, 1, 1},
  {&__pyx_n_s_futures_index, __pyx_k_futures_index, sizeof(__pyx_k_futures_index), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_exchange_pair, __pyx_k_get_exchange_pair, sizeof(__pyx_k_get_exchange_pair), 0, 0, 1, 1},
  {&__pyx_n_s_get_feed_name, __pyx_k_get_feed_name, sizeof(__pyx_k_get_feed_name), 0, 0, 1, 1},
  {&__pyx_kp_u_get_feed_name_not_implemented, __pyx_k_get_feed_name_not_implemented, sizeof(__pyx_k_get_feed_name_not_implemented), 0, 1, 0, 0},
  {&__pyx_n_s_get_pair_from_exchange, __pyx_k_get_pair_from_exchange, sizeof(__pyx_k_get_pair_from_exchange), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_handle_book_adds, __pyx_k_handle_book_adds, sizeof(__pyx_k_handle_book_adds), 0, 0, 1, 1},
  {&__pyx_n_s_high_price, __pyx_k_high_price, sizeof(__pyx_k_high_price), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
  {&__pyx_n_s_interval, __pyx_k_interval, sizeof(__pyx_k_interval), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_kline, __pyx_k_kline, sizeof(__pyx_k_kline), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_leaves_qty, __pyx_k_leaves_qty, sizeof(__pyx_k_leaves_qty), 0, 0, 1, 1},
  {&__pyx_n_s_liquidations, __pyx_k_liquidations, sizeof(__pyx_k_liquidations), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_s_low_price, __pyx_k_low_price, sizeof(__pyx_k_low_price), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_market_info, __pyx_k_market_info, sizeof(__pyx_k_market_info), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_new_event_loop, __pyx_k_new_event_loop, sizeof(__pyx_k_new_event_loop), 0, 0, 1, 1},
  {&__pyx_n_s_new_symbol_separator, __pyx_k_new_symbol_separator, sizeof(__pyx_k_new_symbol_separator), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_commons, __pyx_k_octobot_commons, sizeof(__pyx_k_octobot_commons), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_commons_enums, __pyx_k_octobot_commons_enums, sizeof(__pyx_k_octobot_commons_enums), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_commons_logging, __pyx_k_octobot_commons_logging, sizeof(__pyx_k_octobot_commons_logging), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_commons_symbol_util, __pyx_k_octobot_commons_symbol_util, sizeof(__pyx_k_octobot_commons_symbol_util), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_trading_constants, __pyx_k_octobot_trading_constants, sizeof(__pyx_k_octobot_trading_constants), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_trading_enums, __pyx_k_octobot_trading_enums, sizeof(__pyx_k_octobot_trading_enums), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_trading_exchanges_abstra, __pyx_k_octobot_trading_exchanges_abstra, sizeof(__pyx_k_octobot_trading_exchanges_abstra), 0, 0, 1, 1},
  {&__pyx_n_s_octobot_trading_exchanges_connec, __pyx_k_octobot_trading_exchanges_connec, sizeof(__pyx_k_octobot_trading_exchanges_connec), 0, 0, 1, 1},
  {&__pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_k_octobot_trading_exchanges_connec_2, sizeof(__pyx_k_octobot_trading_exchanges_connec_2), 0, 0, 1, 0},
  {&__pyx_n_s_open_interest, __pyx_k_open_interest, sizeof(__pyx_k_open_interest), 0, 0, 1, 1},
  {&__pyx_n_s_open_price, __pyx_k_open_price, sizeof(__pyx_k_open_price), 0, 0, 1, 1},
  {&__pyx_n_s_order_id, __pyx_k_order_id, sizeof(__pyx_k_order_id), 0, 0, 1, 1},
  {&__pyx_n_s_order_price, __pyx_k_order_price, sizeof(__pyx_k_order_price), 0, 0, 1, 1},
  {&__pyx_n_s_order_size, __pyx_k_order_size, sizeof(__pyx_k_order_size), 0, 0, 1, 1},
  {&__pyx_n_s_pair, __pyx_k_pair, sizeof(__pyx_k_pair), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_price, __pyx_k_price, sizeof(__pyx_k_price), 0, 0, 1, 1},
  {&__pyx_n_s_push_to_channel, __pyx_k_push_to_channel, sizeof(__pyx_k_push_to_channel), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_CryptofeedWebsock, __pyx_k_pyx_unpickle_CryptofeedWebsock, sizeof(__pyx_k_pyx_unpickle_CryptofeedWebsock), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_receipt_timestamp, __pyx_k_receipt_timestamp, sizeof(__pyx_k_receipt_timestamp), 0, 0, 1, 1},
  {&__pyx_n_s_recent_trades, __pyx_k_recent_trades, sizeof(__pyx_k_recent_trades), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_event_loop, __pyx_k_set_event_loop, sizeof(__pyx_k_set_event_loop), 0, 0, 1, 1},
  {&__pyx_n_s_set_logging_level, __pyx_k_set_logging_level, sizeof(__pyx_k_set_logging_level), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_should_uppercase, __pyx_k_should_uppercase, sizeof(__pyx_k_should_uppercase), 0, 0, 1, 1},
  {&__pyx_n_s_side, __pyx_k_side, sizeof(__pyx_k_side), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_symbol, __pyx_k_symbol, sizeof(__pyx_k_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_symbol_separator, __pyx_k_symbol_separator, sizeof(__pyx_k_symbol_separator), 0, 0, 1, 1},
  {&__pyx_n_s_symbol_util, __pyx_k_symbol_util, sizeof(__pyx_k_symbol_util), 0, 0, 1, 1},
  {&__pyx_n_s_symbols, __pyx_k_symbols, sizeof(__pyx_k_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_ticker, __pyx_k_ticker, sizeof(__pyx_k_ticker), 0, 0, 1, 1},
  {&__pyx_n_s_time_frame, __pyx_k_time_frame, sizeof(__pyx_k_time_frame), 0, 0, 1, 1},
  {&__pyx_n_s_timestamp, __pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 0, 1, 1},
  {&__pyx_n_s_trade, __pyx_k_trade, sizeof(__pyx_k_trade), 0, 0, 1, 1},
  {&__pyx_n_s_trades, __pyx_k_trades, sizeof(__pyx_k_trades), 0, 0, 1, 1},
  {&__pyx_n_s_trading_constants, __pyx_k_trading_constants, sizeof(__pyx_k_trading_constants), 0, 0, 1, 1},
  {&__pyx_n_s_trading_enums, __pyx_k_trading_enums, sizeof(__pyx_k_trading_enums), 0, 0, 1, 1},
  {&__pyx_n_s_transactions, __pyx_k_transactions, sizeof(__pyx_k_transactions), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_update_order_book, __pyx_k_update_order_book, sizeof(__pyx_k_update_order_book), 0, 0, 1, 1},
  {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_volume, __pyx_k_volume, sizeof(__pyx_k_volume), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 39, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 62, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":62
 *     @classmethod
 *     def get_feed_name(cls):
 *         raise NotImplementedError("get_feed_name not implemented")             # <<<<<<<<<<<<<<
 * 
 *     def fix_signal_handler(self):
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_get_feed_name_not_implemented); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":61
 * 
 *     @classmethod
 *     def get_feed_name(cls):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_n_s_cls); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_get_feed_name, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 61, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":64
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 *     def fix_signal_handler(self):             # <<<<<<<<<<<<<<
 *         """
 *         Websocket are started in a new thread thus signal handle cannot be used
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_fix_signal_handler, 64, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 64, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":118
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_tuple__21 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_bid, __pyx_n_s_ask, __pyx_n_s_timestamp, __pyx_n_s_receipt_timestamp); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(7, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_ticker, 118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 118, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":128
 *                                        bid_price=None)
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_tuple__22 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_order_id, __pyx_n_s_timestamp, __pyx_n_s_side, __pyx_n_s_amount, __pyx_n_s_price, __pyx_n_s_receipt_timestamp); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(9, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_trade, 128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":143
 *                                        }])
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_tuple__23 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_book, __pyx_n_s_timestamp, __pyx_n_s_receipt_timestamp, __pyx_n_s_book_instance); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(6, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_book, 143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 143, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":162
 *                                        update_order_book=False)
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):             # <<<<<<<<<<<<<<
 *         return [
 *             {
 */
  __pyx_tuple__24 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_book_prices, __pyx_n_s_book_side, __pyx_n_s_order_price, __pyx_n_s_order_size); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_convert_book_prices_to_orders, 162, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 162, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":172
 *         ]
 * 
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,             # <<<<<<<<<<<<<<
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 */
  __pyx_tuple__26 = PyTuple_Pack(17, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_start, __pyx_n_s_stop, __pyx_n_s_interval, __pyx_n_s_trades, __pyx_n_s_open_price, __pyx_n_s_close_price, __pyx_n_s_high_price, __pyx_n_s_low_price, __pyx_n_s_volume, __pyx_n_s_closed, __pyx_n_s_timestamp, __pyx_n_s_receipt_timestamp, __pyx_n_s_time_frame, __pyx_n_s_candle); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(15, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_candle, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 172, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":196
 *                                            candle=candle)
 * 
 *     async def delta(self, feed, symbol, delta, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_delta, __pyx_n_s_timestamp, __pyx_n_s_receipt_timestamp); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_delta, 196, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(0, 196, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":199
 *         pass
 * 
 *     async def liquidations(self, feed, symbol, side, leaves_qty, price, order_id, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_side, __pyx_n_s_leaves_qty, __pyx_n_s_price, __pyx_n_s_order_id, __pyx_n_s_receipt_timestamp); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(8, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_liquidations, 199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 199, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":202
 *         pass
 * 
 *     async def funding(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_funding, 202, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 202, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":205
 *         pass
 * 
 *     async def open_interest(self, feed, symbol, open_interest, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_feed, __pyx_n_s_symbol, __pyx_n_s_open_interest, __pyx_n_s_timestamp, __pyx_n_s_receipt_timestamp); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_open_interest, 205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 205, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":208
 *         pass
 * 
 *     async def volume(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_volume, 208, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 208, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":211
 *         pass
 * 
 *     async def futures_index(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_futures_index, 211, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 211, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":214
 *         pass
 * 
 *     async def market_info(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Coingecko only
 * 
 */
  __pyx_tuple__33 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_market_info, 214, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 214, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":217
 *         pass  # Coingecko only
 * 
 *     async def transactions(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Whale alert only
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_transactions, 217, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 217, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":220
 *         pass  # Whale alert only
 * 
 *     def get_pair_from_exchange(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */
  __pyx_tuple__35 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pair); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_get_pair_from_exchange, 220, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 220, __pyx_L1_error)

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":232
 *         return ""
 * 
 *     def get_exchange_pair(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */
  __pyx_tuple__37 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pair); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_octobot_trading_exchanges_connec_2, __pyx_n_s_get_exchange_pair, 232, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 232, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__39 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)
 */
  __pyx_tuple__41 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(2, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_CryptofeedWebsocketConnector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__43 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CryptofeedWebsock, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_236055009 = PyInt_FromLong(236055009L); if (unlikely(!__pyx_int_236055009)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.exchanges.abstract_websocket_exchange"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange = __Pyx_ImportType(__pyx_t_1, "octobot_trading.exchanges.abstract_websocket_exchange", "AbstractWebsocketExchange", sizeof(struct __pyx_obj_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange = (struct __pyx_vtabstruct_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange*)__Pyx_GetVtable(__pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange->tp_dict); if (unlikely(!__pyx_vtabptr_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector = &__pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.__pyx_base = *__pyx_vtabptr_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.get_pair_from_exchange = (PyObject *(*)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *, int __pyx_skip_dispatch))__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_pair_from_exchange;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.get_exchange_pair = (PyObject *(*)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *, int __pyx_skip_dispatch))__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_get_exchange_pair;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.start = (void (*)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *))__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_start;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.subscribe_feeds = (void (*)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *))__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_feeds;
  __pyx_vtable_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.subscribe_candle_feed = (void (*)(struct __pyx_obj_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector *, PyObject *))__pyx_f_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_subscribe_candle_feed;
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_base = __pyx_ptype_15octobot_trading_9exchanges_27abstract_websocket_exchange_AbstractWebsocketExchange;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector.tp_dict, __pyx_vtabptr_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CryptofeedWebsocketConnector, (PyObject *)&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct__ticker;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_1_trade;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_2_book;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle) < 0) __PYX_ERR(0, 172, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_3_candle;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_4_delta;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_5_liquidations;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_6_funding;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_7_open_interest;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_8_volume;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_9_futures_index;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_10_market_info;
  if (PyType_Ready(&__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions.tp_dictoffset && __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions = &__pyx_type_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector___pyx_scope_struct_11_transactions;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.util.initializable"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_15octobot_trading_4util_13initializable_Initializable = __Pyx_ImportType(__pyx_t_1, "octobot_trading.util.initializable", "Initializable", sizeof(struct __pyx_obj_15octobot_trading_4util_13initializable_Initializable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_15octobot_trading_4util_13initializable_Initializable) __PYX_ERR(3, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.exchanges.abstract_exchange"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange = __Pyx_ImportType(__pyx_t_1, "octobot_trading.exchanges.abstract_exchange", "AbstractExchange", sizeof(struct __pyx_obj_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange) __PYX_ERR(4, 20, __pyx_L1_error)
  __pyx_vtabptr_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange = (struct __pyx_vtabstruct_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange*)__Pyx_GetVtable(__pyx_ptype_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange->tp_dict); if (unlikely(!__pyx_vtabptr_15octobot_trading_9exchanges_17abstract_exchange_AbstractExchange)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.exchanges.exchange_config_data"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig = __Pyx_ImportType(__pyx_t_1, "octobot_trading.exchanges.exchange_config_data", "ExchangeConfig", sizeof(struct __pyx_obj_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig) __PYX_ERR(5, 19, __pyx_L1_error)
  __pyx_vtabptr_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig = (struct __pyx_vtabstruct_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig*)__Pyx_GetVtable(__pyx_ptype_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig->tp_dict); if (unlikely(!__pyx_vtabptr_15octobot_trading_9exchanges_20exchange_config_data_ExchangeConfig)) __PYX_ERR(5, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.exchanges.exchange_manager"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager = __Pyx_ImportType(__pyx_t_1, "octobot_trading.exchanges.exchange_manager", "ExchangeManager", sizeof(struct __pyx_obj_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager) __PYX_ERR(6, 24, __pyx_L1_error)
  __pyx_vtabptr_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager = (struct __pyx_vtabstruct_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager*)__Pyx_GetVtable(__pyx_ptype_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager->tp_dict); if (unlikely(!__pyx_vtabptr_15octobot_trading_9exchanges_16exchange_manager_ExchangeManager)) __PYX_ERR(6, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("octobot_trading.util.config_util"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "is_trader_enabled", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_is_trader_enabled, "int (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "is_trader_simulator_enabled", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_is_trader_simulator_enabled, "int (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "is_currency_enabled", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_is_currency_enabled, "int (PyObject *, PyObject *, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "is_trade_history_loading_enabled", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled, "int (PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_is_trade_history_loading_enabled *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_symbols", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_get_symbols, "PyObject *(PyObject *, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_all_currencies", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_get_all_currencies, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_all_currencies *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_pairs", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_get_pairs, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_pairs *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_market_pair", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_get_market_pair, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_15octobot_trading_4util_11config_util_get_market_pair *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_reference_market", (void (**)(void))&__pyx_f_15octobot_trading_4util_11config_util_get_reference_market, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcryptofeed_websocket_connector(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcryptofeed_websocket_connector(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cryptofeed_websocket_connector(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cryptofeed_websocket_connector(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cryptofeed_websocket_connector(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cryptofeed_websocket_connector' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cryptofeed_websocket_connector(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cryptofeed_websocket_connector", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_octobot_trading__exchanges__connectors__cryptofeed_websocket_connector) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector")) {
      if (unlikely(PyDict_SetItemString(modules, "octobot_trading.exchanges.connectors.cryptofeed_websocket_connector", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":16
 * #  You should have received a copy of the GNU Lesser General Public
 * #  License along with this library.
 * import asyncio             # <<<<<<<<<<<<<<
 * import logging
 * 
 */
  __pyx_t_1 = __Pyx_patch_asyncio(__Pyx_Import(__pyx_n_s_asyncio, 0, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":17
 * #  License along with this library.
 * import asyncio
 * import logging             # <<<<<<<<<<<<<<
 * 
 * import cryptofeed
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_logging, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":19
 * import logging
 * 
 * import cryptofeed             # <<<<<<<<<<<<<<
 * import cryptofeed.callback as cryptofeed_callbacks
 * import cryptofeed.defines as cryptofeed_constants
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_cryptofeed, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cryptofeed, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":20
 * 
 * import cryptofeed
 * import cryptofeed.callback as cryptofeed_callbacks             # <<<<<<<<<<<<<<
 * import cryptofeed.defines as cryptofeed_constants
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__15);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_cryptofeed_callback, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cryptofeed_callbacks, __pyx_t_2) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":21
 * import cryptofeed
 * import cryptofeed.callback as cryptofeed_callbacks
 * import cryptofeed.defines as cryptofeed_constants             # <<<<<<<<<<<<<<
 * 
 * import octobot_commons
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__15);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_cryptofeed_defines, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cryptofeed_constants, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":23
 * import cryptofeed.defines as cryptofeed_constants
 * 
 * import octobot_commons             # <<<<<<<<<<<<<<
 * import octobot_commons.enums as commons_enums
 * import octobot_commons.logging as commons_logging
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_octobot_commons, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_octobot_commons, __pyx_t_1) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":24
 * 
 * import octobot_commons
 * import octobot_commons.enums as commons_enums             # <<<<<<<<<<<<<<
 * import octobot_commons.logging as commons_logging
 * import octobot_commons.symbol_util as symbol_util
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__15);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_octobot_commons_enums, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_commons_enums, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":25
 * import octobot_commons
 * import octobot_commons.enums as commons_enums
 * import octobot_commons.logging as commons_logging             # <<<<<<<<<<<<<<
 * import octobot_commons.symbol_util as symbol_util
 * import octobot_trading.constants as trading_constants
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__15);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_octobot_commons_logging, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_commons_logging, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":26
 * import octobot_commons.enums as commons_enums
 * import octobot_commons.logging as commons_logging
 * import octobot_commons.symbol_util as symbol_util             # <<<<<<<<<<<<<<
 * import octobot_trading.constants as trading_constants
 * import octobot_trading.enums as trading_enums
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__15);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_octobot_commons_symbol_util, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_symbol_util, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":27
 * import octobot_commons.logging as commons_logging
 * import octobot_commons.symbol_util as symbol_util
 * import octobot_trading.constants as trading_constants             # <<<<<<<<<<<<<<
 * import octobot_trading.enums as trading_enums
 * import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__15);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_octobot_trading_constants, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trading_constants, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":28
 * import octobot_commons.symbol_util as symbol_util
 * import octobot_trading.constants as trading_constants
 * import octobot_trading.enums as trading_enums             # <<<<<<<<<<<<<<
 * import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
 * from octobot_trading.enums import WebsocketFeeds as Feeds
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__15);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_octobot_trading_enums, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trading_enums, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":29
 * import octobot_trading.constants as trading_constants
 * import octobot_trading.enums as trading_enums
 * import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket             # <<<<<<<<<<<<<<
 * from octobot_trading.enums import WebsocketFeeds as Feeds
 * from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__15);
  __Pyx_GIVEREF(__pyx_n_s__15);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__15);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_octobot_trading_exchanges_abstra, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_abstract_websocket, __pyx_t_1) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":30
 * import octobot_trading.enums as trading_enums
 * import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
 * from octobot_trading.enums import WebsocketFeeds as Feeds             # <<<<<<<<<<<<<<
 * from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_WebsocketFeeds);
  __Pyx_GIVEREF(__pyx_n_s_WebsocketFeeds);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_WebsocketFeeds);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_octobot_trading_enums, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_WebsocketFeeds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Feeds, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":31
 * import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket
 * from octobot_trading.enums import WebsocketFeeds as Feeds
 * from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ExchangeConstantsOrderBookInfoCo);
  __Pyx_GIVEREF(__pyx_n_s_ExchangeConstantsOrderBookInfoCo);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ExchangeConstantsOrderBookInfoCo);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_octobot_trading_enums, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ExchangeConstantsOrderBookInfoCo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ECOBIC, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":35
 * 
 * class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
 *     LOGGERS = ["feedhandler"]             # <<<<<<<<<<<<<<
 *     CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_feedhandler);
  __Pyx_GIVEREF(__pyx_n_u_feedhandler);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_feedhandler);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_LOGGERS, __pyx_t_1) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":36
 * class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
 *     LOGGERS = ["feedhandler"]
 *     CRYPTOFEED_DEFAULT_MARKET_SEPARATOR = "-"             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, config: object, exchange_manager: object):
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_CRYPTOFEED_DEFAULT_MARKET_SEPARA, __pyx_kp_u__16) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":61
 * 
 *     @classmethod
 *     def get_feed_name(cls):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_3get_feed_name, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_get, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_get_feed_name, __pyx_t_1) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":60
 *         }
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def get_feed_name(cls):
 *         raise NotImplementedError("get_feed_name not implemented")
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector, __pyx_n_s_get_feed_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_get_feed_name, __pyx_t_2) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":64
 *         raise NotImplementedError("get_feed_name not implemented")
 * 
 *     def fix_signal_handler(self):             # <<<<<<<<<<<<<<
 *         """
 *         Websocket are started in a new thread thus signal handle cannot be used
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_5fix_signal_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_fix, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_fix_signal_handler, __pyx_t_2) < 0) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":118
 *             self.logger.error(f"Failed to start websocket feed : {e}")
 * 
 *     async def ticker(self, feed, symbol, bid, ask, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_7ticker, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_tic, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_ticker, __pyx_t_2) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":128
 *                                        bid_price=None)
 * 
 *     async def trade(self, feed, symbol, order_id, timestamp, side, amount, price, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_10trade, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_tra, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_trade, __pyx_t_2) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":143
 *                                        }])
 * 
 *     async def book(self, feed, symbol, book, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         if symbol:
 *             symbol = self.get_pair_from_exchange(symbol)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_13book, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_boo, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_book, __pyx_t_2) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":162
 *                                        update_order_book=False)
 * 
 *     def _convert_book_prices_to_orders(self, book_prices, book_side):             # <<<<<<<<<<<<<<
 *         return [
 *             {
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_16_convert_book_prices_to_orders, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector__co, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_convert_book_prices_to_orders, __pyx_t_2) < 0) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":172
 *         ]
 * 
 *     async def candle(self, feed, symbol, start, stop, interval, trades, open_price, close_price, high_price,             # <<<<<<<<<<<<<<
 *                      low_price, volume, closed, timestamp, receipt_timestamp):
 *         if symbol:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_18candle, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_can, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__5)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_candle, __pyx_t_2) < 0) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":196
 *                                            candle=candle)
 * 
 *     async def delta(self, feed, symbol, delta, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_21delta, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_del, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_delta, __pyx_t_2) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":199
 *         pass
 * 
 *     async def liquidations(self, feed, symbol, side, leaves_qty, price, order_id, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_24liquidations, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_liq, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_liquidations, __pyx_t_2) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":202
 *         pass
 * 
 *     async def funding(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_27funding, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_fun, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__8)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_funding, __pyx_t_2) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":205
 *         pass
 * 
 *     async def open_interest(self, feed, symbol, open_interest, timestamp, receipt_timestamp):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_30open_interest, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_ope, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__9)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_open_interest, __pyx_t_2) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":208
 *         pass
 * 
 *     async def volume(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_33volume, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_vol, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_volume, __pyx_t_2) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":211
 *         pass
 * 
 *     async def futures_index(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_36futures_index, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_fut, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_futures_index, __pyx_t_2) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":214
 *         pass
 * 
 *     async def market_info(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Coingecko only
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_39market_info, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_mar, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_market_info, __pyx_t_2) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":217
 *         pass  # Coingecko only
 * 
 *     async def transactions(self, **kwargs):             # <<<<<<<<<<<<<<
 *         pass  # Whale alert only
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_42transactions, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_tra_2, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_transactions, __pyx_t_2) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":220
 *         pass  # Whale alert only
 * 
 *     def get_pair_from_exchange(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_45get_pair_from_exchange, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_get_2, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_get_pair_from_exchange, __pyx_t_2) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":232
 *         return ""
 * 
 *     def get_exchange_pair(self, pair) -> str:             # <<<<<<<<<<<<<<
 *         try:
 *             return symbol_util.convert_symbol(
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_47get_exchange_pair, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector_get_3, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_get_exchange_pair, __pyx_t_2) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_49__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector___r, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CryptofeedWebsocketConnector, (type(self), 0xe11e9e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CryptofeedWebsocketConnector__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_28CryptofeedWebsocketConnector_51__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CryptofeedWebsocketConnector___s, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(2, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_CryptofeedWebsocketConnector);

  /* "(tree fragment)":1
 * def __pyx_unpickle_CryptofeedWebsocketConnector(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_15octobot_trading_9exchanges_10connectors_30cryptofeed_websocket_connector_1__pyx_unpickle_CryptofeedWebsocketConnector, 0, __pyx_n_s_pyx_unpickle_CryptofeedWebsock, NULL, __pyx_n_s_octobot_trading_exchanges_connec, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CryptofeedWebsock, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "octobot_trading/exchanges/connectors/cryptofeed_websocket_connector.py":1
 * #  Drakkar-Software OctoBot-Trading             # <<<<<<<<<<<<<<
 * #  Copyright (c) Drakkar-Software, All rights reserved.
 * #
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init octobot_trading.exchanges.connectors.cryptofeed_websocket_connector", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init octobot_trading.exchanges.connectors.cryptofeed_websocket_connector");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Coroutine */
static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
    PyObject_GC_UnTrack(self);
    Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
    PyObject_GC_Del(self);
}
static int __Pyx_CoroutineAwait_traverse(__pyx_CoroutineAwaitObject *self, visitproc visit, void *arg) {
    Py_VISIT(self->coroutine);
    return 0;
}
static int __Pyx_CoroutineAwait_clear(__pyx_CoroutineAwaitObject *self) {
    Py_CLEAR(self->coroutine);
    return 0;
}
static PyObject *__Pyx_CoroutineAwait_Next(__pyx_CoroutineAwaitObject *self) {
    return __Pyx_Generator_Next(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_Send(__pyx_CoroutineAwaitObject *self, PyObject *value) {
    return __Pyx_Coroutine_Send(self->coroutine, value);
}
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args) {
    return __Pyx_Coroutine_Throw(self->coroutine, args);
}
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_self(PyObject *self) {
    Py_INCREF(self);
    return self;
}
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_CoroutineAwait_no_new(CYTHON_UNUSED PyTypeObject *type, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwargs) {
    PyErr_SetString(PyExc_TypeError, "cannot instantiate type, use 'await coroutine' instead");
    return NULL;
}
#endif
static PyMethodDef __pyx_CoroutineAwait_methods[] = {
    {"send", (PyCFunction) __Pyx_CoroutineAwait_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_CoroutineAwait_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_CoroutineAwait_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
    {0, 0, 0, 0}
};
static PyTypeObject __pyx_CoroutineAwaitType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    (destructor) __Pyx_CoroutineAwait_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("A wrapper object implementing __await__ for coroutines."),
    (traverseproc) __Pyx_CoroutineAwait_traverse,
    (inquiry) __Pyx_CoroutineAwait_clear,
    0,
    0,
    __Pyx_CoroutineAwait_self,
    (iternextfunc) __Pyx_CoroutineAwait_Next,
    __pyx_CoroutineAwait_methods,
    0                         ,
    0                      ,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if !CYTHON_COMPILING_IN_PYPY
    __Pyx_CoroutineAwait_no_new,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
#if PY_VERSION_HEX < 0x030500B1 || defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
    __pyx_CoroutineAwaitObject *await = PyObject_GC_New(__pyx_CoroutineAwaitObject, __pyx_CoroutineAwaitType);
    if (unlikely(!await)) return NULL;
    Py_INCREF(coroutine);
    await->coroutine = coroutine;
    PyObject_GC_Track(await);
    return (PyObject*)await;
}
#endif
#if PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_await_method(PyObject *coroutine, CYTHON_UNUSED PyObject *arg) {
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static PyObject *__Pyx_Coroutine_await(PyObject *coroutine) {
    if (unlikely(!coroutine || !__Pyx_Coroutine_Check(coroutine))) {
        PyErr_SetString(PyExc_TypeError, "invalid input, expected coroutine");
        return NULL;
    }
    return __Pyx__Coroutine_await(coroutine);
}
#endif
static PyObject *
__Pyx_Coroutine_get_frame(CYTHON_UNUSED __pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    Py_RETURN_NONE;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_compare(PyObject *obj, PyObject *other, int op) {
    PyObject* result;
    switch (op) {
        case Py_EQ: result = (other == obj) ? Py_True : Py_False; break;
        case Py_NE: result = (other != obj) ? Py_True : Py_False; break;
        default:
            result = Py_NotImplemented;
    }
    Py_INCREF(result);
    return result;
}
#endif
static PyMethodDef __pyx_Coroutine_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
#if PY_VERSION_HEX < 0x030500B1
    {"__await__", (PyCFunction) __Pyx_Coroutine_await_method, METH_NOARGS,
     (char*) PyDoc_STR("__await__() -> return an iterator to be used in await expression.")},
#endif
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Coroutine_memberlist[] = {
    {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being awaited, or None")},
    {(char*) "cr_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Coroutine_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the coroutine"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the coroutine"), 0},
    {(char *) "cr_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_ASYNC_SLOTS
static __Pyx_PyAsyncMethodsStruct __pyx_Coroutine_as_async = {
    __Pyx_Coroutine_await,
    0,
    0,
};
#endif
static PyTypeObject __pyx_CoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    0,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Coroutine_init(void) {
    __pyx_CoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_CoroutineType = __Pyx_FetchCommonType(&__pyx_CoroutineType_type);
    if (unlikely(!__pyx_CoroutineType))
        return -1;
#ifdef __Pyx_IterableCoroutine_USED
    if (unlikely(__pyx_IterableCoroutine_init() == -1))
        return -1;
#endif
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonType(&__pyx_CoroutineAwaitType_type);
    if (unlikely(!__pyx_CoroutineAwaitType))
        return -1;
    return 0;
}

/* GetAwaitIter */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o) {
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(o)) {
        return __Pyx_NewRef(o);
    }
#endif
    return __Pyx__Coroutine_GetAwaitableIter(o);
}
static void __Pyx_Coroutine_AwaitableIterError(PyObject *source) {
#if PY_VERSION_HEX >= 0x030600B3 || defined(_PyErr_FormatFromCause)
    _PyErr_FormatFromCause(
        PyExc_TypeError,
        "'async for' received an invalid object "
        "from __anext__: %.100s",
        Py_TYPE(source)->tp_name);
#elif PY_MAJOR_VERSION >= 3
    PyObject *exc, *val, *val2, *tb;
    assert(PyErr_Occurred());
    PyErr_Fetch(&exc, &val, &tb);
    PyErr_NormalizeException(&exc, &val, &tb);
    if (tb != NULL) {
        PyException_SetTraceback(val, tb);
        Py_DECREF(tb);
    }
    Py_DECREF(exc);
    assert(!PyErr_Occurred());
    PyErr_Format(
        PyExc_TypeError,
        "'async for' received an invalid object "
        "from __anext__: %.100s",
        Py_TYPE(source)->tp_name);
    PyErr_Fetch(&exc, &val2, &tb);
    PyErr_NormalizeException(&exc, &val2, &tb);
    Py_INCREF(val);
    PyException_SetCause(val2, val);
    PyException_SetContext(val2, val);
    PyErr_Restore(exc, val2, tb);
#else
    source++;
#endif
}
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *obj) {
    PyObject *res;
#if CYTHON_USE_ASYNC_SLOTS
    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(obj);
    if (likely(am && am->am_await)) {
        res = (*am->am_await)(obj);
    } else
#endif
#if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
    if (PyCoro_CheckExact(obj)) {
        return __Pyx_NewRef(obj);
    } else
#endif
#if CYTHON_COMPILING_IN_CPYTHON && defined(CO_ITERABLE_COROUTINE)
    if (PyGen_CheckExact(obj) && ((PyGenObject*)obj)->gi_code && ((PyCodeObject *)((PyGenObject*)obj)->gi_code)->co_flags & CO_ITERABLE_COROUTINE) {
        return __Pyx_NewRef(obj);
    } else
#endif
    {
        PyObject *method = NULL;
        int is_method = __Pyx_PyObject_GetMethod(obj, __pyx_n_s_await, &method);
        if (likely(is_method)) {
            res = __Pyx_PyObject_CallOneArg(method, obj);
        } else if (likely(method)) {
            res = __Pyx_PyObject_CallNoArg(method);
        } else
            goto slot_error;
        Py_DECREF(method);
    }
    if (unlikely(!res)) {
        __Pyx_Coroutine_AwaitableIterError(obj);
        goto bad;
    }
    if (unlikely(!PyIter_Check(res))) {
        PyErr_Format(PyExc_TypeError,
                     "__await__() returned non-iterator of type '%.100s'",
                     Py_TYPE(res)->tp_name);
        Py_CLEAR(res);
    } else {
        int is_coroutine = 0;
        #ifdef __Pyx_Coroutine_USED
        is_coroutine |= __Pyx_Coroutine_Check(res);
        #endif
        #if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
        is_coroutine |= PyCoro_CheckExact(res);
        #endif
        if (unlikely(is_coroutine)) {
            /* __await__ must return an *iterator*, not
               a coroutine or another awaitable (see PEP 492) */
            PyErr_SetString(PyExc_TypeError,
                            "__await__() returned a coroutine");
            Py_CLEAR(res);
        }
    }
    return res;
slot_error:
    PyErr_Format(PyExc_TypeError,
                 "object %.100s can't be used in 'await' expression",
                 Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* CoroutineYieldFrom */
static PyObject* __Pyx__Coroutine_Yield_From_Generic(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
    if (unlikely(!source_gen)) {
        return NULL;
    }
    if (__Pyx_Coroutine_Check(source_gen)) {
        retval = __Pyx_Generator_Next(source_gen);
    } else {
#if CYTHON_USE_TYPE_SLOTS
        retval = Py_TYPE(source_gen)->tp_iternext(source_gen);
#else
        retval = PyIter_Next(source_gen);
#endif
    }
    if (retval) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    if (__Pyx_Coroutine_Check(source)) {
        if (unlikely(((__pyx_CoroutineObject*)source)->yieldfrom)) {
            PyErr_SetString(
                PyExc_RuntimeError,
                "coroutine is being awaited already");
            return NULL;
        }
        retval = __Pyx_Generator_Next(source);
#ifdef __Pyx_AsyncGen_USED
    } else if (__pyx_PyAsyncGenASend_CheckExact(source)) {
        retval = __Pyx_async_gen_asend_iternext(source);
#endif
    } else {
        return __Pyx__Coroutine_Yield_From_Generic(gen, source);
    }
    if (retval) {
        Py_INCREF(source);
        gen->yieldfrom = source;
    }
    return retval;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PatchInspect */
static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* PatchAsyncIO */
static PyObject* __Pyx_patch_asyncio(PyObject* module) {
#if PY_VERSION_HEX < 0x030500B2 &&\
        (defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED)) &&\
        (!defined(CYTHON_PATCH_ASYNCIO) || CYTHON_PATCH_ASYNCIO)
    PyObject *patch_module = NULL;
    static int asyncio_patched = 0;
    if (unlikely((!asyncio_patched) && module)) {
        PyObject *package;
        package = __Pyx_Import(__pyx_n_s_asyncio_coroutines, NULL, 0);
        if (package) {
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "coroutines"), ""
"try:\n"
"    coro_types = _module._COROUTINE_TYPES\n"
"except AttributeError: pass\n"
"else:\n"
"    if _cython_coroutine_type is not None and _cython_coroutine_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_coroutine_type,)\n"
"    if _cython_generator_type is not None and _cython_generator_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_generator_type,)\n"
"_module._COROUTINE_TYPES = coro_types\n"
            );
        } else {
            PyErr_Clear();
            package = __Pyx_Import(__pyx_n_s_asyncio_tasks, NULL, 0);
            if (unlikely(!package)) goto asyncio_done;
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "tasks"), ""
"if hasattr(_module, 'iscoroutine'):\n"
"    old_types = getattr(_module.iscoroutine, '_cython_coroutine_types', None)\n"
"    if old_types is None or not isinstance(old_types, set):\n"
"        old_types = set()\n"
"        def cy_wrap(orig_func, type=type, cython_coroutine_types=old_types):\n"
"            def cy_iscoroutine(obj): return type(obj) in cython_coroutine_types or orig_func(obj)\n"
"            cy_iscoroutine._cython_coroutine_types = cython_coroutine_types\n"
"            return cy_iscoroutine\n"
"        _module.iscoroutine = cy_wrap(_module.iscoroutine)\n"
"    if _cython_coroutine_type is not None:\n"
"        old_types.add(_cython_coroutine_type)\n"
"    if _cython_generator_type is not None:\n"
"        old_types.add(_cython_generator_type)\n"
            );
        }
        Py_DECREF(package);
        if (unlikely(!patch_module)) goto ignore;
asyncio_done:
        PyErr_Clear();
        asyncio_patched = 1;
#ifdef __Pyx_Generator_USED
        {
            PyObject *inspect_module;
            if (patch_module) {
                inspect_module = PyObject_GetAttr(patch_module, __pyx_n_s_inspect);
                Py_DECREF(patch_module);
            } else {
                inspect_module = __Pyx_Import(__pyx_n_s_inspect, NULL, 0);
            }
            if (unlikely(!inspect_module)) goto ignore;
            inspect_module = __Pyx_patch_inspect(inspect_module);
            if (unlikely(!inspect_module)) {
                Py_DECREF(module);
                module = NULL;
            }
            Py_XDECREF(inspect_module);
        }
#else
        if ((0)) return __Pyx_patch_inspect(module);
#endif
    }
    return module;
ignore:
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch asyncio package with custom generator type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    if ((0)) return __Pyx_patch_inspect(__Pyx_Coroutine_patch_module(module, NULL));
#endif
    return module;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
